# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from functools import wraps

import numpy as np
from tensortrax.math import linalg, log, special, sqrt
from tensortrax.math import sum as sum1
from tensortrax.math import trace

det = linalg.det
inv = linalg.inv
eigvalsh = linalg.eigvalsh

from_triu_1d = special.from_triu_1d
triu_1d = special.triu_1d


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient."""

    @wraps(fun)
    def apply_iso(C, *args, **kwargs):
        return fun(det(C) ** (-1 / 3) * C, *args, **kwargs)

    return apply_iso


def saint_venant_kirchhoff(C, mu, lmbda):
    r"""Strain energy function of the isotropic hyperelastic
    `Saint-Venant Kirchhoff <https://en.wikipedia.org/wiki/Hyperelastic_material#Saint_Venant-Kirchhoff_model>`_
    material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : float
        Second Lamé constant (shear modulus).
    lmbda : float
        First Lamé constant (shear modulus).

    Notes
    -----
    ..  math::

        \psi = \mu I_2 + \lambda \frac{I_1^2}{2}

    With the first and second invariant of the Green-Lagrange strain tensor
    :math:`\boldsymbol{E} = \frac{1}{2} (\boldsymbol{C} - \boldsymbol{1})`.

    ..  math::

        I_1 &= \text{tr}\left( \boldsymbol{E} \right)

        I_2 &= \boldsymbol{E} : \boldsymbol{E}

    Examples
    --------
    ..  warning::
        The Saint-Venant Kirchhoff material formulation is unstable for large strains.

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.saint_venant_kirchhoff, mu=1.0, lmbda=20.0)
    >>> ax = umat.plot(incompressible=False)
    UserWarning: Uniaxial data with volume ratio det(F) <= 0 included.
    UserWarning: Planar Shear data with volume ratio det(F) <= 0 included.
    UserWarning: Biaxial data with volume ratio det(F) <= 0 included.

    ..  image:: images/umat_saint_venant_kirchhoff.png
        :width: 400px

    """
    I1 = trace(C) / 2 - 3 / 2
    I2 = trace(C @ C) / 4 - trace(C) / 2 + 3 / 4
    return mu * I2 + lmbda * I1**2 / 2


def neo_hooke(C, mu):
    r"""Strain energy function of the isotropic hyperelastic
    `Neo-Hookean <https://en.wikipedia.org/wiki/Neo-Hookean_solid>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : float
        Shear modulus.

    Notes
    -----
    ..  math::

        \psi = \frac{\mu}{2} \left(\text{tr}\left(\hat{\boldsymbol{C}}\right) - 3\right)

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.neo_hooke, mu=1.0)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_neo_hooke.png
        :width: 400px

    """
    return mu / 2 * (det(C) ** (-1 / 3) * trace(C) - 3)


def mooney_rivlin(C, C10, C01):
    r"""Strain energy function of the isotropic hyperelastic
    `Mooney-Rivlin <https://en.wikipedia.org/wiki/Mooney-Rivlin_solid>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C10 : float
        First material parameter associated to the first invariant.
    C01 : float
        Second material parameter associated to the second invariant.

    Notes
    -----
    ..  math::

        \psi = C_{10} \left(\hat{I}_1 - 3 \right) + C_{01} \left(\hat{I}_2 - 3 \right)

    With the first and second main invariant of the distortional part of the right
    Cauchy-Green deformation tensor.

    ..  math::

        \hat{I}_1 &= J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

        \hat{I}_2 &= J^{-4/3} \frac{1}{2} \left(
            \text{tr}\left(\boldsymbol{C}\right)^2 -
            \text{tr}\left(\boldsymbol{C}^2\right)
        \right)

    The doubled sum of both material parameters is equal to the shear modulus
    :math:`\mu`.

    ..  math::

        \mu = 2 \left( C_{10} + C_{01} \right)

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.mooney_rivlin, C10=0.3, C01=0.8)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_mooney_rivlin.png
        :width: 400px

    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (I1**2 - J3**2 * trace(C @ C)) / 2
    return C10 * (I1 - 3) + C01 * (I2 - 3)


def yeoh(C, C10, C20, C30):
    r"""Strain energy function of the isotropic hyperelastic
    `Yeoh <https://en.wikipedia.org/wiki/Yeoh_(hyperelastic_model)>`_
    material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C10 : float
        Material parameter associated to the linear term of the first invariant.
    C20 : float
        Material parameter associated to the quadratic term of the first invariant.
    C30 : float
        Material parameter associated to the cubic term of the first invariant.

    Notes
    -----
    ..  math::

        \psi = C_{10} \left(\hat{I}_1 - 3 \right) + C_{20} \left(\hat{I}_1 - 3 \right)^2
             + C_{30} \left(\hat{I}_1 - 3 \right)^3

    With the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor.

    ..  math::

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    The :math:`C_{10}` material parameter is equal to half the initial shear modulus
    :math:`\mu`.

    ..  math::

        \mu = 2 C_{10}

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.yeoh, C10=0.5, C20=-0.1, C30=0.02)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_yeoh.png
        :width: 400px

    """

    I1 = det(C) ** (-1 / 3) * trace(C)
    return C10 * (I1 - 3) + C20 * (I1 - 3) ** 2 + C30 * (I1 - 3) ** 3


def third_order_deformation(C, C10, C01, C11, C20, C30):
    r"""Strain energy function of the isotropic hyperelastic
    `Third-Order-Deformation <https://onlinelibrary.wiley.com/doi/abs/10.1002/app.1975.070190723>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C10 : float
        Material parameter associated to the linear term of the first invariant.
    C01 : float
        Material parameter associated to the linear term of the second invariant.
    C11 : float
        Material parameter associated to the mixed term of the first and second
        invariant.
    C20 : float
        Material parameter associated to the quadratic term of the first invariant.
    C30 : float
        Material parameter associated to the cubic term of the first invariant.

    Notes
    -----
    ..  math::

        \psi &= C_{10} \left(\hat{I}_1 - 3 \right) + C_{01} \left(\hat{I}_2 - 3 \right)
             + C_{11} \left(\hat{I}_1 - 3 \right) \left(\hat{I}_2 - 3 \right)

            &+ C_{20} \left(\hat{I}_1 - 3 \right)^2
             + C_{30} \left(\hat{I}_1 - 3 \right)^3

    With the first and second main invariant of the distortional part of the right
    Cauchy-Green deformation tensor.

    ..  math::

        \hat{I}_1 &= J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

        \hat{I}_2 &= J^{-4/3} \frac{1}{2} \left(
            \text{tr}\left(\boldsymbol{C}\right)^2 -
            \text{tr}\left(\boldsymbol{C}^2\right)
        \right)

    The doubled sum of the material parameters :math:`C_{10}` and :math:`C_{01}` is
    equal to the initial shear modulus :math:`\mu`.

    ..  math::

        \mu = 2 \left( C_{10} + C_{01} \right)

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(
    >>>     fem.third_order_deformation, C10=0.5, C01=0.1, C11=0.01, C20=-0.1, C30=0.02
    >>> )
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_third_order_deformation.png
        :width: 400px

    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (I1**2 - J3**2 * trace(C @ C)) / 2
    return (
        C10 * (I1 - 3)
        + C01 * (I2 - 3)
        + C11 * (I1 - 3) * (I2 - 3)
        + C20 * (I1 - 3) ** 2
        + C30 * (I1 - 3) ** 3
    )


def ogden(C, mu, alpha):
    r"""Strain energy function of the isotropic hyperelastic
    `Ogden <https://en.wikipedia.org/wiki/Ogden_(hyperelastic_model)>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : list of float
        List of moduli.
    alpha : list of float
        List of stretch exponents.

    Notes
    -----
    ..  math::

        \psi = \sum_i \frac{2 \mu_i}{\alpha^2_i} \left(
            \lambda_1^{\alpha_i} + \lambda_2^{\alpha_i} + \lambda_3^{\alpha_i} - 3
        \right)

    The sum of the moduli :math:`\mu_i` is equal to the initial shear modulus
    :math:`\mu`.

    ..  math::

        \mu = \sum_i \mu_i

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.ogden, mu=[1, 0.2], alpha=[1.7, -1.5])
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_ogden.png
        :width: 400px

    """

    wC = det(C) ** (-1 / 3) * eigvalsh(C)
    return sum1([2 * m / a**2 * (sum1(wC ** (a / 2)) - 3) for m, a in zip(mu, alpha)])


def arruda_boyce(C, C1, limit):
    r"""Strain energy function of the isotropic hyperelastic
    `Arruda-Boyce <https://en.wikipedia.org/wiki/Arruda-Boyce_model>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C1 : list of float
        Initial shear modulus.
    limit : list of float
        Limiting stretch at which the polymer chain network becomes locked 
        :math:`\lambda_m`.

    Notes
    -----
    ..  math::

        \psi = C_1 \sum_{i=1}^5 \alpha_i \beta^{i-1} \left( \hat{I}_1^i - 3^i \right)

    With the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor
    
    ..  math::
    
        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)
    
    and :math_`\alpha_i` and :math`\beta`.
    
    ..  math::
        
        \boldsymbol{\alpha} &= \begin{bmatrix} 
            \frac{1}{2} \\ 
            \frac{1}{20} \\
            \frac{11}{1050} \\
            \frac{19}{7000} \\
            \frac{519}{673750}
        \end{bmatrix}
        
        \beta &= \frac{1}{\lambda_m^2}
    
    The initial shear modulus is a function of both material parameters.
    
    ..  math::
        
        \mu = C_1 \left( 
            1 + \frac{3}{5 \lambda_m^2} + \frac{99}{175 \lambda_m^4} 
              + \frac{513}{875 \lambda_m^6} + \frac{42039}{67375 \lambda_m^8} 
        \right)

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.arruda_boyce, C1=1.0, limit=3.2)
    >>> ax = umat.plot(incompressible=True)
    
    ..  image:: images/umat_arruda_boyce.png
        :width: 400px

    """
    I1 = det(C) ** (-1 / 3) * trace(C)

    alphas = [1 / 2, 1 / 20, 11 / 1050, 19 / 7000, 519 / 673750]
    beta = 1 / limit**2

    out = []
    for j, alpha in enumerate(alphas):
        i = j + 1
        out.append(alpha * beta ** (i - 1) * (I1**i - 3**i))

    return C1 * sum1(out)


def extended_tube(C, Gc, delta, Ge, beta):
    r"""Strain energy function of the isotropic hyperelastic
    `Extended Tube <https://www.doi.org/10.5254/1.3538822>`_ [1] material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    Gc : float
        Cross-link contribution to the initial shear modulus.
    delta : float
         Finite extension parameter of the polymer strands.
    Ge : float
        Constraint contribution to the initial shear modulus.
    beta : float
        Global rearrangements of cross-links upon deformation (release of topological
        constraints).

    Notes
    -----
    ..  math::

        \psi = \frac{G_c}{2} \left[ \frac{\left( 1 - \delta^2 \right)
            \left( \hat{I}_1 - 3 \right)}{1 - \delta^2 \left( \hat{I}_1 - 3 \right)} +
            \ln \left( 1 - \delta^2 \left( \hat{I}_1 - 3 \right) \right) \right] +
            \frac{2 G_e}{\beta^2} \left( \hat{\lambda}_1^{-\beta} +
            \hat{\lambda}_2^{-\beta} + \hat{\lambda}_3^{-\beta} - 3 \right)

    With the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor

    ..  math::

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    and the principal stretches, obtained from the distortional part of the right
    Cauchy-Green deformation tensor

    ..  math::

        \hat{\lambda}_\alpha = J^{-1/3} \lambda_\alpha

    The initial shear modulus results from the sum of the cross-link and the constraint
    contributions to the total initial shear modulus.

    ..  math::

        \mu = G_e + G_c

    Examples
    --------

    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(
    >>>     fem.extended_tube, Gc=0.1867, Ge=0.2169, beta=0.2, delta=0.09693
    >>> )
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_extended_tube.png
        :width: 400px

    References
    ----------
    [1] M. Kaliske and G. Heinrich, "An Extended Tube-Model for Rubber Elasticity:
    Statistical-Mechanical Theory and Finite Element Implementation", Rubber Chemistry
    and Technology, vol. 72, no. 4. Rubber Division, ACS, pp. 602–632, Sep. 01, 1999.
    doi: 10.5254/1.3538822.

    """
    J3 = det(C) ** (-1 / 3)
    D = J3 * trace(C)
    wC = J3 * eigvalsh(C)
    g = (1 - delta**2) * (D - 3) / (1 - delta**2 * (D - 3))
    Wc = Gc / 2 * (g + log(1 - delta**2 * (D - 3)))
    We = 2 * Ge / beta**2 * sum1(wC ** (-beta / 2) - 1)
    return Wc + We


def van_der_waals(C, mu, limit, a, beta):
    """Strain energy function of the Van der Waals material formulation.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.van_der_waals, mu=1.0, beta=0.1, a=0.5, limit=5.0)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_van_der_waals.png
        :width: 400px
    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (trace(C) ** 2 - J3**2 * trace(C @ C)) / 2
    Im = (1 - beta) * I1 + beta * I2
    Im.x[np.isclose(Im.x, 3)] += 1e-8
    eta = sqrt((Im - 3) / (limit**2 - 3))
    return mu * (
        -(limit**2 - 3) * (log(1 - eta) + eta) - 2 / 3 * a * ((Im - 3) / 2) ** (3 / 2)
    )


@isochoric_volumetric_split
def finite_strain_viscoelastic(C, Cin, mu, eta, dtime):
    """Finite strain viscoelastic material formulation.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(
    >>>     fem.finite_strain_viscoelastic, mu=1.0, eta=1.0, dtime=1.0, nstatevars=6
    >>> )
    >>> ax = umat.plot(
    >>>    incompressible=True,
    >>>    ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
    >>>    ps=None,
    >>>    bx=None,
    >>> )

    ..  image:: images/umat_finite_strain_viscoelastic.png
        :width: 400px
    """

    # update of state variables by evolution equation
    Ci = from_triu_1d(Cin, like=C) + mu / eta * dtime * C
    Ci = det(Ci) ** (-1 / 3) * Ci

    # first invariant of elastic part of right Cauchy-Green deformation tensor
    I1 = trace(C @ inv(Ci))

    # strain energy function and state variable
    return mu / 2 * (I1 - 3), triu_1d(Ci)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..math import sym as symmetric
from ._base import Field


class FieldAxisymmetric(Field):
    r"""An axisymmetric :class:`~felupe.Field` on points of a two-dimensional
    :class:`~felupe.Region` with dimension ``dim`` (default is 2) and initial point
    ``values`` (default is 0).
    
    Parameters
    ----------
    region : Region
        The region on which the field will be created.
    dim : int, optional
        The dimension of the field (default is 2).
    values : float or array, optional
        A single value for all components of the field or an array of
        shape (region.mesh.npoints, dim)`. Default is 0.0.
    
    Notes
    -----
    * component 1 =  axial component
    * component 2 = radial component

    ..  code-block::

         x_2 (radial direction)

          ^
          |        _
          |       / \
        --|-----------------> x_1 (axial rotation axis)
                  \_^

    This is a modified :class:`Field` in which the radial coordinates are evaluated at
    the numeric integration points ``q`` for each cell ``c``. The :meth:`grad`-method is
    modified in such a way that it does not only contain the in-plane 2d-gradient but
    also the circumferential stretch, see Eq. :eq:`gradient_axi`.
    
    ..  math::
        :label: gradient_axi
        
        \frac{\partial \boldsymbol{u}}{\partial \boldsymbol{X}} = 
            \begin{bmatrix}
                \left( 
                    \frac{\partial \boldsymbol{u}}{\partial \boldsymbol{X}} 
                \right)_{2d} & \boldsymbol{0} \\
                \boldsymbol{0}^T & \frac{u_r}{R}
            \end{bmatrix}
    
    See Also
    --------
    felupe.Field : Field on points of a :class:`~felupe.Region` with dimension ``dim``
       and initial point ``values``.

    """

    def __init__(self, region, dim=2, values=0.0):
        # init base Field
        super().__init__(region, dim=dim, values=values)

        # create scalar-valued field of radial point values
        self.scalar = Field(region, dim=1, values=region.mesh.points[:, 1])

        # interpolate radial point values to integration points of each cell
        # in the region
        self.radius = self.scalar.interpolate()

    def _interpolate_2d(self, out=None):
        """Interpolate 2D field values at points and evaluate them at the
        integration points of all cells in the region."""

        # interpolated field values "aI"
        # evaluated at quadrature point "q"
        # for cell "c"
        return np.einsum(
            "ca...,aqc->...qc",
            self.values[self.region.mesh.cells],
            self.region.h,
            out=out,
        )

    def interpolate(self, out=None):
        # out-argument is not supported
        # if out is not None:
        #     out = out[:2]

        # extend dimension of in-plane 2d-gradient
        return np.pad(self._interpolate_2d(out=None), ((0, 1), (0, 0), (0, 0)))

    def _grad_2d(self, sym=False, out=None):
        """In-plane 2D gradient as partial derivative of field values at points
        w.r.t. the undeformed coordinates, evaluated at the integration points
        of all cells in the region. Optionally, the symmetric part of the
        gradient is returned.

        Arguments
        ---------
        sym : bool, optional (default is False)
            Calculate the symmetric part of the gradient.
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).

        Returns
        -------
        ndarray
            In-plane 2D-gradient as partial derivative of field values at points
            w.r.t. undeformed coordinates, evaluated at the integration points
            of all cells in the region.
        """

        # gradient as partial derivative of field component "I" at point "a"
        # w.r.t. undeformed coordinate "J" evaluated at quadrature point "q"
        # for each cell "c"
        g = np.einsum(
            "ca...,aJqc->...Jqc",
            self.values[self.region.mesh.cells],
            self.region.dhdX,
            out=out,
        )

        if sym:
            return symmetric(g, out=g)
        else:
            return g

    def grad(self, sym=False, out=None):
        """3D-gradient as partial derivative of field values at points w.r.t.
        the undeformed coordinates, evaluated at the integration points of all
        cells in the region. Optionally, the symmetric part of the gradient is
        returned.

        ..  code-block::

                        |  dudX(2d) :   0   |
            dudX(axi) = | ..................|
                        |     0     : u_r/R |

        Arguments
        ---------
        sym : bool, optional
            Calculate the symmetric part of the gradient (default is False).
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).

        Returns
        -------
        ndarray
            Full 3D-gradient as partial derivative of field values at points
            w.r.t. undeformed coordinates, evaluated at the integration points
            of all cells in the region.
        """

        # out-argument is not supported
        # if out is not None:
        #     out = out[:2, :2]

        # extend dimension of in-plane 2d-gradient
        g = np.pad(self._grad_2d(sym=sym, out=None), ((0, 1), (0, 1), (0, 0), (0, 0)))

        # set dudX_33 = u_r / R
        g[-1, -1] = self.interpolate()[1] / self.radius

        return g



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix as sparsematrix
from scipy.sparse.linalg import spsolve

from ..assembly import IntegralFormCartesian
from ..element import Tetra, Triangle, QuadraticTriangle, QuadraticTetra
from ..field import Field
from ..quadrature import Tetrahedron as TetrahedronQuadrature
from ..quadrature import Triangle as TriangleQuadrature
from ..region import Region


def topoints(values, region, sym=True, mode="tensor"):
    "Shift of scalar or tensorial values at quadrature points to mesh-points."

    rows = region.mesh.cells.T.ravel()
    cols = np.zeros_like(rows)

    if mode == "tensor":
        dim = values.shape[0]
        if sym:
            if dim == 3:
                ij = [(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)]
            elif dim == 2:
                ij = [(0, 0), (1, 1), (0, 1)]
        else:
            if dim == 3:
                ij = [
                    (0, 0),
                    (0, 1),
                    (0, 2),
                    (1, 0),
                    (1, 1),
                    (1, 2),
                    (2, 0),
                    (2, 1),
                    (2, 2),
                ]
            elif dim == 2:
                ij = [(0, 0), (0, 1), (1, 0), (1, 1)]

        out = Field(region, dim=len(ij)).values

        for a, (i, j) in enumerate(ij):
            out[:, a] = (
                sparsematrix(
                    (values.reshape(dim, dim, -1)[i, j], (rows, cols)),
                    shape=(region.mesh.npoints, 1),
                ).toarray()[:, 0]
                / region.mesh.cells_per_point
            )

    elif mode == "scalar":
        out = sparsematrix(
            (values.ravel(), (rows, cols)), shape=(region.mesh.npoints, 1)
        ).toarray()[:, 0]
        out = out / region.mesh.cells_per_point

    return out


def extrapolate(values, region, average=True, mean=False):
    """Extrapolate (and optionally averaging) scalar or vectorial values
    at quadrature points to mesh-points by inversion.

    Parameters
    ----------
    values : ndarray of shape (..., q, c)
        Array with values located at the quadrature points ``q`` of cells ``c``.
    region : Region
        A region used to extrapolate the values to the mesh-points.
    average : bool, optional
        A flag to return values averaged at mesh-points if True and the mesh of the
        region is not already disconnected or to return values on a disconnected mesh
        with discontinuous values at cell transitions if False (default is True).
    mean : bool, optional
        A flag to take the cell-means, averaged by the quadrature weights, of the
        values (default is False).

    Returns
    -------
    Field.values : ndarray of shape (p, ...)
        Array of values projected to the mesh-points ``p``.
    """

    dim = values.shape[:-2]
    size = int(np.prod(dim))
    weights = region.quadrature.weights

    # transpose values
    idx = np.arange(len(values.shape))
    idx[: len(dim)] = idx[: len(dim)][::-1]
    values = values.transpose(idx)

    if mean:
        # evaluate how often the values must be repeated to match the number
        # of element-points
        reps = np.ones(len(values.shape), dtype=int)
        reps[-2] = len(region.element.points)

        # np.average(keepdims=True) requires numpy >= 1.23.0
        values = np.average(values, axis=-2, weights=weights)

        # workaround for np.average(keepdims=True)
        shape = values.shape
        shape = np.insert(shape, -1, 1)
        values = values.reshape(*shape)

        # broadcast averaged values to match the number of element-points
        values = np.broadcast_to(values, shape=np.broadcast_shapes(shape, reps))

    # reshape from (shape, nint.points, nelements) to 1d-values
    u = values.T.reshape(-1, size)

    # disconnect the mesh
    m = region.mesh.disconnect()

    if mean:
        # region on disconnected mesh with original quadrature scheme
        # a single-point quadrature would be sufficient
        # but would require additional (not available) informations
        r = Region(m, region.element, region.quadrature, grad=False)
    else:
        # region on disconnected mesh with inverse quadrature scheme
        r = Region(m, region.element, region.quadrature.inv(), grad=False)

    # field for values on disconnected mesh; project values to mesh-points
    f = Field(r, dim=size, values=u)
    v = f.interpolate()

    if mean:
        # due to the usage of the original quadrature scheme the averaging must be
        # applied again
        # np.average(keepdims=True) requires numpy >= 1.23.0
        v = np.average(v, axis=-2, weights=weights).reshape(size, 1, -1)

        # broadcast averaged values to match the number of element-points
        shape = np.array([*v.shape[:-2], len(region.element.points), v.shape[-1]])
        v = np.broadcast_to(v, shape=shape)

    if average:
        # create dummy field for values on original mesh
        # (used for calculation of sparse-matrix indices)
        g = Field(region, dim=size)

        # average values
        w = sparsematrix(
            (v.T.ravel(), g.indices.ai), shape=(size * region.mesh.npoints, 1)
        ).toarray().reshape(-1, size) / region.mesh.cells_per_point.reshape(-1, 1)

    else:
        w = v.T.reshape(-1, size)

    return w


def project(values, region, average=True, mean=False, dV=None, solver=spsolve):
    """Project given values at quadrature-points projected to mesh-points.

    Parameters
    ----------
    values : ndarray of shape (..., q, c)
        Array with values located at the quadrature points ``q`` of cells ``c``.
    region : Region
        A region used to project the values to the mesh-points.
    average : bool, optional
        A flag to return values averaged at mesh-points if True and the mesh of the
        region is not already disconnected or to return values on a disconnected mesh
        with discontinuous values at cell transitions if False (default is True).
    mean : bool, optional
        A flag to take the cell-means of the values. Quadrature weights are ignored.
        Default is False.
    dV : ndarray of shape (q, c) or None, optional
        Differential volumes located at the quadrature points of the cells. If None,
        the differential volumes are taken from the region (default is None).
    solver : callable, optional
        A function for a sparse solver with signature ``x=solver(A, b)``. Default is
        :func:`scipy.sparse.linalg.spsolve`.

    Returns
    -------
    Field.values : ndarray of shape (p, ...)
        Array of values projected to the mesh-points ``p``.

    Notes
    -----
    The projection finds :math:`\hat{\boldsymbol{u}}`, located at mesh-points ``p``, for
    values :math:`v`, evaluated at quadrature-points ``q``, such that the variation of
    the minimization problem in Eq. :eq:`project` is solved.

    ..  math::
        :label: project

        \Pi &= \int_V \frac{1}{2} (v - u)^2 \delta u\ dV \quad \rightarrow \quad \min

        \delta \Pi &= \int_V (v - u) \delta u\ dV = 0

    This leads to assembled system-matrix and -vector(s) as shown in
    Eq. :eq:`project-forms`

    ..  math::
        :label: project-forms

        &\int_V v\ \delta u\ dV \qquad \rightarrow \qquad \hat{\boldsymbol{A}}

        &\int_V u\ \delta u\ dV \qquad \rightarrow \qquad \hat{\boldsymbol{b}}

    of an equation system to be solved, see Eq. :eq:`project-solve`. The right-arrows
    in Eq. :eq:`project-forms` represent the assembly if the integral forms into the
    system matrix or vectors.

    ..  math::
        :label: project-solve

        \hat{\boldsymbol{A}}\ \hat{\boldsymbol{u}} = \hat{\boldsymbol{b}}

    The region, on which the values are projected to, may contain a mesh with a
    different (e.g. a disconnected) points- but same cells-array. The only requirement
    on the region is that it must use a compatible quadrature scheme. With the values at
    mesh-points at hand, new fields may be created to project between fields on
    different regions. If the array of differential volumes is not given, then the
    region must be created with ``Region(grad=True)`` to include the differential
    volumes.

    ..  warning::

        Triangular element formulations require regions with quadrature schemes with
        higher integration orders, the default choices are not sufficient for
        projection. For :class:`~felupe.RegionTriangle` and :class:`~felupe.RegionTetra`
        a second-order scheme is used.

        ..  code-block::

            import felupe as fem

            mesh = fem.Rectangle(n=2).triangulate().add_midpoints_edges()
            quadrature = fem.TriangleQuadrature(order=5)
            fem.RegionQuadraticTriangle(mesh, quadrature=quadrature)

        For :class:`~felupe.RegionQuadraticTriangle` and
        :class:`~felupe.RegionQuadraticTetra` use a fifth-order scheme.

        ..  code-block::

            mesh = fem.Cube(n=2).triangulate().add_midpoints_edges()
            quadrature = fem.TetrahedronQuadrature(order=5)
            fem.RegionQuadraticTetra(mesh, quadrature=quadrature)

    Examples
    --------

    ..  code-block::

        import felupe as fem

        mesh = fem.Rectangle(n=2).triangulate()
        region = fem.RegionTriangle(mesh)
        field = fem.FieldAxisymmetric(region, dim=2)
        values = field.extract()

        region2 = region.copy()
        region2.reload(quadrature=fem.TriangleQuadrature(order=2))
        values_projected = project(values, region2)
    """

    mesh = None
    if not average:
        mesh = region.mesh.disconnect()  # for non-continuous results

    if mean:
        values = np.mean(values, axis=-2, keepdims=True)  # cell-based means

    # quadrature schemes for projection
    # triangles and tetrahedrons require quadratic quadratures for projection
    element = None
    quadrature = None
    if values.shape[-2] == 1:
        scheme = {
            Triangle: TriangleQuadrature(order=2),
            Tetra: TetrahedronQuadrature(order=2),
            QuadraticTriangle: TriangleQuadrature(order=5),
            QuadraticTetra: TetrahedronQuadrature(order=5),
        }
        quadrature = scheme.get(type(region.element))

    # copy and reload the region if necessary
    if mesh is not None or element is not None or quadrature is not None:
        region = region.copy(mesh, element, quadrature)

    shape = values.shape[:-2]  # tensor-components
    size = np.prod(shape).astype(int)  # tensor-size (enforce int for empty tuple)
    axes = values.shape[-2:]  # trailing axes

    if dV is None:
        dV = region.dV  # take the differential volumes from the region to project

    # lhs (volume matrix) is constant for all items
    v = u = Field(region, dim=1)  # 1d-field for lhs
    A = IntegralFormCartesian(np.ones((1, 1)), v=v, dV=dV, u=u).assemble()

    # fix diagonal items of the matrix for points not connected to cells
    zeros_on_diagonal = A.diagonal() == 0
    if np.any(zeros_on_diagonal):
        A = A.tolil()
        A[zeros_on_diagonal, zeros_on_diagonal] = 1
        A = A.tocsr()

    # field of unknowns with projected values
    x = Field(region, dim=size)

    # solve with individual right-hand-sides (per tensor-component of values)
    b = IntegralFormCartesian(values.reshape(size, *axes), v=x, dV=dV).assemble()
    x.values[:] = spsolve(A, b.toarray().reshape(-1, size)).reshape(x.values.shape)

    return x.values.reshape(-1, *shape)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt


def identity(A=None, dim=None, shape=None):
    r"""Return identity matrices with ones on the diagonal of the first two axes and
    zeros elsewhere. 
    
    Parameters
    ----------
    A : ndarray of shape (N, M, ...) or None, optional
        The array of input matrices. If provided, the dimension and the shape of the
        trailing axes are taken from this array. If None, ``dim`` and ``shape`` are
        required. Default is None.
    dim : int or None, optional
        The dimension of the matrix axes. If None, it is taken from ``A`` (default is
        None).
    shape : tuple of int or None, optional
        A tuple containing the shape of the trailing axes (batch dimensions). Default is
        None.
    
    Returns
    -------
    ndarray of shape (N, M, *np.ones_like(...)) or (dim, dim, *np.ones_like(...))
        The identity matrix.
    
    Notes
    -----
    The first two axes are the matrix dimensions and all remaining trailing axes are
    treated as batch dimensions. As the identity matrix is idependent of the input
    matrices ``A``, the size of the trailing axes is reduced to one.
    
    The identity matrix is defined by Eq. :eq:`math-identity`
    
    ..  math::
        :label: math-identity
        
        \boldsymbol{1} = \boldsymbol{A} \boldsymbol{A}^{-1} 
            = \boldsymbol{A}^{-1} \boldsymbol{A}
    
    and it is shown in Eq. :eq:`math-identity-items`.
    
    ..  math::
        :label: math-identity-items
        
        \boldsymbol{1} = \begin{bmatrix}
                1 & 0 & 0 \\
                0 & 1 & 0 \\
                0 & 0 & 1
            \end{bmatrix}
    
    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>> 
    >>> A = np.random.rand(3, 2, 8, 20)
    >>> I = fem.math.identity(A)
    >>> I.shape
    (3, 3, 1, 1)
    
    With given dimension of the matrix axes the shape of the output is different.
    
    >>> fem.math.identity(A, dim=2).shape
    (2, 2, 1, 1)
    
    Note how the number of batch axes change if a ``shape`` is given.
    
    >>> fem.math.identity(A, shape=(4, 7, 3)).shape
    (3, 3, 1, 1, 1)
    
    See Also
    --------
    numpy.eye : Return a 2-D array with ones on the diagonal and zeros elsewhere.
    """

    M = None
    if A is not None:
        N, M = A.shape[:2]
        shapeA = A.shape[2:]
        if dim is None:
            dim = N
        if shape is None:
            shape = shapeA

    ones = (1,) * len(shape)
    eye = np.eye(N=dim, M=M)
    return eye.reshape(*eye.shape, *ones)


def sym(A, out=None):
    r"""Return the symmetric parts of second-order tensors.

    Parameters
    ----------
    A : ndarray of shape (M, M, ...)
        The array of second-order tensors.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (M, M, ...)
        Array with the symmetric parts of the second-order tensors.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The symmetric part of a second-order tensor is obtained by Eq. :eq:`math-symmetric`.

    ..  math::
        :label: math-symmetric

        \text{sym} \left( \boldsymbol{A} \right) = \frac{1}{2} \left(
                \boldsymbol{A} + \boldsymbol{A}^T
            \right)

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(18, dtype=float).reshape(2, 3, 3).T)
    >>> A[..., 0]
    array([[0., 1., 2.],
           [3., 4., 5.],
           [6., 7., 8.]])

    >>> fem.math.sym(A)[..., 0]
    array([[0., 2., 4.],
           [2., 4., 6.],
           [4., 6., 8.]])

    """
    out = np.add(A, transpose(A), out=out)
    return np.multiply(out, 0.5, out=out)


def dya(A, B, mode=2, parallel=False, **kwargs):
    r"""Return the dyadic product of two first-order or two second-order tensors.

    Parameters
    ----------
    A : ndarray of shape (N, ...) or (N, N, ...)
        Array with first-order or second-order tensors.
    B : ndarray of shape (M, ...) or (M, M, ...)
        Array with first-order or second-order tensors.
    mode : int, optional
        Mode of operation. Return the dyadic products of two second-order tensors with
        2 and the dyadic products of two first-order tensors with 1. Default is 2.
    parallel : bool, optional
        A flag to enable a threaded evaluation of the results (default is False).
    **kwargs : dict, optional
        Optional keyword-arguments for :func:`numpy.multiply`, e.g. ``out=None``.

    Returns
    -------
    ndarray of shape (N, M, ...) or (N, N, M, M, ...)
        The array of dyadic products.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions. The definition of the dyadic product is given in Eq.
    :eq:`math-dya2` for two second-order tensors

    ..  math::
        :label: math-dya2

        \mathbb{C} &= \boldsymbol{A} \otimes \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{ij}\ B_{kl}

    and in Eq. :eq:`math-dya1` for two first-order tensors.

    ..  math::
        :label: math-dya1

        \boldsymbol{C} &= \boldsymbol{a} \otimes \boldsymbol{b}

        C_{ij} &= a_{i} \otimes b_{j}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> B = fem.math.transpose(np.arange(100, 109, dtype=float).reshape(1, 3, 3).T)
    >>> C = fem.math.dya(A, B)
    >>> C.shape
    (3, 3, 3, 3, 1)

    >>> C[..., 0].reshape(9, 9)
    array([[  0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.],
           [100., 101., 102., 103., 104., 105., 106., 107., 108.],
           [200., 202., 204., 206., 208., 210., 212., 214., 216.],
           [300., 303., 306., 309., 312., 315., 318., 321., 324.],
           [400., 404., 408., 412., 416., 420., 424., 428., 432.],
           [500., 505., 510., 515., 520., 525., 530., 535., 540.],
           [600., 606., 612., 618., 624., 630., 636., 642., 648.],
           [700., 707., 714., 721., 728., 735., 742., 749., 756.],
           [800., 808., 816., 824., 832., 840., 848., 856., 864.]])

    See Also
    --------
    felupe.math.cdya : Crossed-dyadic product of two second-order tensors.
    felupe.math.cdya_ik : ik-crossed dyadic product of two second-order tensors.
    felupe.math.cdya_il : il-crossed dyadic product of two second-order tensors.

    """

    if mode == 2:
        return np.multiply(A[:, :, None, None], B[None, None, :, :], **kwargs)
    elif mode == 1:
        return np.multiply(A[:, None], B[None, :], **kwargs)
    else:
        raise ValueError("unknown mode. (1 or 2)", mode)


def inv(A, determinant=None, full_output=False, sym=False, out=None):
    r"""Return the inverses of second-order tensors.

    Parameters
    ----------
    A : ndarray of shape (1, 1, ...), (2, 2, ...) or (3, 3, ...)
        The array of second-order tensors.
    determinant : ndarray or None, optional
        The array with the pre-evaluated determinants of the second-order tensors (
        default is None).
    full_output : bool, optional
        A flag to return the array of inverses and the determinants (default is False).
    sym : bool, optional
        A flag to assume symmetric second-order tensors. Only the upper-triangle
        elements are taken into account. Default is False.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (1, 1, ...), (2, 2, ...) or (3, 3, ...)
        The inverses of second-order tensors.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The inverse of a three-dimensional second-order tensor is obtained by Eq.
    :eq:`math-inv`

    ..  math::
        :label: math-inv

        \boldsymbol{A}^{-1} \boldsymbol{A} &= \boldsymbol{A} \boldsymbol{A}^{-1} =
            \boldsymbol{1}

        \boldsymbol{A}^{-1} &= \frac{1}{\det(\boldsymbol{A})}
            \text{cof}(\boldsymbol{A})^T

        \boldsymbol{A}^{-1} &= \frac{1}{\det(\boldsymbol{A})} \begin{bmatrix}
                \left( \boldsymbol{A}_2 \times \boldsymbol{A}_3 \right)^T \\
                \left( \boldsymbol{A}_3 \times \boldsymbol{A}_1 \right)^T \\
                \left( \boldsymbol{A}_1 \times \boldsymbol{A}_2 \right)^T
            \end{bmatrix}

    with the column (grid) vectors :math:`\boldsymbol{A}_j`, see Eq. :eq:`math-gridvec`.

    ..  math::
        :label: math-gridvec

        \boldsymbol{A} = \begin{bmatrix}
                \boldsymbol{A}_1 & \boldsymbol{A}_2 & \boldsymbol{A}_3
            \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3, 1)
    >>> A[..., 0]
    array([[1. , 1.3, 1.5],
           [1.3, 1.1, 1.4],
           [1.5, 1.4, 1.2]])
    
    >>> invA = fem.math.inv(A)
    >>> invA[..., 0]
    array([[-2.01892744,  1.70347003,  0.5362776 ],
           [ 1.70347003, -3.31230284,  1.73501577],
           [ 0.5362776 ,  1.73501577, -1.86119874]])
    
    >>> fem.math.dot(A, invA)[..., 0].round(3)
    array([[ 1., -0.,  0.],
           [-0.,  1.,  0.],
           [-0., -0.,  1.]])

    See Also
    --------
    felupe.math.det : Return the determinants of second order tensors.
    felupe.math.cof : Return the cofactors of second order tensors.
    """

    detAinvA = out
    if detAinvA is None:
        detAinvA = np.zeros_like(A)

    x1 = None
    x2 = None
    if A.shape[:2] == (3, 3):
        # diagonal items
        x1 = np.multiply(A[1, 2], A[2, 1], out=x1)
        x2 = np.multiply(A[1, 1], A[2, 2], out=x2)
        np.add(-x1, x2, out=detAinvA[0, 0])

        x1 = np.multiply(A[0, 2], A[2, 0], out=x1)
        x2 = np.multiply(A[0, 0], A[2, 2], out=x2)
        np.add(-x1, x2, out=detAinvA[1, 1])

        x1 = np.multiply(A[0, 1], A[1, 0], out=x1)
        x2 = np.multiply(A[0, 0], A[1, 1], out=x2)
        np.add(-x1, x2, out=detAinvA[2, 2])

        # upper-triangle off-diagonal
        x1 = np.multiply(A[0, 1], A[2, 2], out=x1)
        x2 = np.multiply(A[0, 2], A[2, 1], out=x2)
        np.add(-x1, x2, out=detAinvA[0, 1])

        x1 = np.multiply(A[0, 2], A[1, 1], out=x1)
        x2 = np.multiply(A[0, 1], A[1, 2], out=x2)
        np.add(-x1, x2, out=detAinvA[0, 2])

        x1 = np.multiply(A[0, 0], A[1, 2], out=x1)
        x2 = np.multiply(A[0, 2], A[1, 0], out=x2)
        np.add(-x1, x2, out=detAinvA[1, 2])

        if sym:
            detAinvA[1, 0] = detAinvA[0, 1]
            detAinvA[2, 0] = detAinvA[0, 2]
            detAinvA[2, 1] = detAinvA[1, 2]
        else:
            # lower-triangle off-diagonal
            x1 = np.multiply(A[1, 0], A[2, 2], out=x1)
            x2 = np.multiply(A[2, 0], A[1, 2], out=x2)
            np.add(-x1, x2, out=detAinvA[1, 0])

            x1 = np.multiply(A[2, 0], A[1, 1], out=x1)
            x2 = np.multiply(A[1, 0], A[2, 1], out=x2)
            np.add(-x1, x2, out=detAinvA[2, 0])

            x1 = np.multiply(A[0, 0], A[2, 1], out=x1)
            x2 = np.multiply(A[2, 0], A[0, 1], out=x2)
            np.add(-x1, x2, out=detAinvA[2, 1])

    elif A.shape[:2] == (2, 2):
        detAinvA[0, 0] = A[1, 1]
        detAinvA[0, 1] = -A[0, 1]
        detAinvA[1, 0] = -A[1, 0]
        detAinvA[1, 1] = A[0, 0]

    elif A.shape[:2] == (1, 1):
        detAinvA[0, 0] = 1

    else:
        raise ValueError(
            " ".join(
                [
                    "Wrong shape of first two axes.",
                    "Must be (1, 1), (2, 2) or (3, 3) but {A.shape[:2]} is given.",
                ]
            )
        )

    if determinant is None:
        detA = det(A, out=x1)
    else:
        detA = determinant

    if full_output:
        return np.divide(detAinvA, detA, out=detAinvA), detA
    else:
        return np.divide(detAinvA, detA, out=detAinvA)


def det(A, out=None):
    r"""Return the determinants of second order tensors.

    Parameters
    ----------
    A : ndarray of shape (M, M, ...)
        The array of second-order tensors.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (...)
        Array with the determinants.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The determinant of a second-order tensor is obtained by Eq. :eq:`math-det`.

    ..  math::
        :label: math-det

        \det \left( \boldsymbol{A} \right) &=
              A_{11}~A_{22}~A_{33}
            + A_{12}~A_{23}~A_{31}
            + A_{13}~A_{21}~A_{32}

           &- A_{31}~A_{22}~A_{13}
            - A_{32}~A_{23}~A_{11}
            - A_{33}~A_{21}~A_{12}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T) / 10
    >>> A += np.eye(3).reshape(3, 3, 1)
    >>> A[..., 0]
    >>> A.shape
    (3, 3, 1)

    >>> A[..., 0]
    array([[1. , 0.1, 0.2],
           [0.3, 1.4, 0.5],
           [0.6, 0.7, 1.8]])

    >>> fem.math.det(A)[..., 0]
    array(2.02)

    """

    detA = out
    if detA is None:
        detA = np.zeros_like(A[0, 0])
    else:
        detA.fill(0)

    if A.shape[:2] == (3, 3):
        tmp = np.multiply(A[0, 0], A[1, 1])
        np.multiply(tmp, A[2, 2], out=tmp)
        np.add(detA, tmp, out=detA)

        tmp = np.multiply(A[0, 1], A[1, 2], out=tmp)
        np.multiply(tmp, A[2, 0], out=tmp)
        np.add(detA, tmp, out=detA)

        tmp = np.multiply(A[0, 2], A[1, 0], out=tmp)
        np.multiply(tmp, A[2, 1], out=tmp)
        np.add(detA, tmp, out=detA)

        tmp = np.multiply(A[2, 0], A[1, 1], out=tmp)
        np.multiply(tmp, A[0, 2], out=tmp)
        np.add(detA, -tmp, out=detA)

        tmp = np.multiply(A[2, 1], A[1, 2], out=tmp)
        np.multiply(tmp, A[0, 0], out=tmp)
        np.add(detA, -tmp, out=detA)

        tmp = np.multiply(A[2, 2], A[1, 0], out=tmp)
        np.multiply(tmp, A[0, 1], out=tmp)
        np.add(detA, -tmp, out=detA)

    elif A.shape[:2] == (2, 2):
        tmp = np.multiply(A[0, 0], A[1, 1])
        np.add(detA, tmp, out=detA)

        tmp = np.multiply(A[1, 0], A[0, 1], out=tmp)
        np.add(detA, -tmp, out=detA)

    elif A.shape[:2] == (1, 1):
        np.add(detA, A[0, 0], out=detA)

    else:
        raise ValueError(
            " ".join(
                [
                    "Wrong shape of first two axes.",
                    "Must be (1, 1), (2, 2) or (3, 3) but {A.shape[:2]} is given.",
                ]
            )
        )
    return detA


def dev(A, out=None):
    r"""Return the deviatoric parts of second-order tensors.

    Parameters
    ----------
    A : ndarray of shape (M, M, ...)
        The array of second-order tensors.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (M, M, ...)
        The deviatoric parts of second-order tensors.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The deviatoric part of a three-dimensional second-order tensor is obtained by Eq.
    :eq:`math-dev`.

    ..  math::
        :label: math-dev

        \text{dev} \left( \boldsymbol{A} \right) = \boldsymbol{A}
            - \frac{\text{tr}(\boldsymbol{A})}{3} \boldsymbol{1}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> A[..., 0]
    array([[0., 1., 2.],
           [3., 4., 5.],
           [6., 7., 8.]])

    >>> fem.math.sym(A)[..., 0]
    array([[0., 2., 4.],
           [2., 4., 6.],
           [4., 6., 8.]])

    See Also
    --------
    numpy.eye : Return a 2-D array with ones on the diagonal and zeros elsewhere.
    """
    dim = A.shape[0]
    return np.add(A, -trace(A) / dim * identity(A), out=out)


def cof(A, sym=False, out=None):
    r"""Return the cofactors of second-order tensors.

    Parameters
    ----------
    A : ndarray of shape (1, 1, ...), (2, 2, ...) or (3, 3, ...)
        The array of second-order tensors.
    sym : bool, optional
        A flag to assume symmetric second-order tensors. Only the upper-triangle
        elements are taken into account. Default is False.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (1, 1, ...), (2, 2, ...) or (3, 3, ...)
        The cofactors of second-order tensors.

    Notes
    -----
    The first two axes are the matrix dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The inverse of a three-dimensional second-order tensor is obtained by Eq.
    :eq:`math-cof`

    ..  math::
        :label: math-cof

        \text{cof}(\boldsymbol{A}) &= \det (\boldsymbol{A}) \boldsymbol{A}^{-T}

        \text{cof}(\boldsymbol{A}) &= \begin{bmatrix}
                \boldsymbol{A}_2 \times \boldsymbol{A}_3 &
                \boldsymbol{A}_3 \times \boldsymbol{A}_1 &
                \boldsymbol{A}_1 \times \boldsymbol{A}_2
            \end{bmatrix}

    with the column (grid) vectors :math:`\boldsymbol{A}_j`, see Eq. :eq:`math-gridv`.

    ..  math::
        :label: math-gridv

        \boldsymbol{A} = \begin{bmatrix}
                \boldsymbol{A}_1 & \boldsymbol{A}_2 & \boldsymbol{A}_3
            \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3, 1)
    >>> A[..., 0]
    array([[1. , 1.3, 1.5],
           [1.3, 1.1, 1.4],
           [1.5, 1.4, 1.2]])

    >>> cofA = fem.math.cof(A)
    >>> cofA[..., 0]
    array([[-0.64,  0.54,  0.17],
           [ 0.54, -1.05,  0.55],
           [ 0.17,  0.55, -0.59]])

    >>> (fem.math.det(A) * fem.math.transpose(fem.math.inv(A)))[..., 0]
    array([[-0.64,  0.54,  0.17],
           [ 0.54, -1.05,  0.55],
           [ 0.17,  0.55, -0.59]])

    See Also
    --------
    felupe.math.det : Return the determinants of second order tensors.
    felupe.math.inv : Return the inverses of second order tensors.
    """
    return transpose(inv(A, determinant=1.0, sym=sym, out=out))


def eig(A, eig=np.linalg.eig):
    "Eigenvalues and -vectors of matrix A."
    wA, vA = eig(A.T)
    return wA.T, vA.T


def eigh(A):
    "Eigenvalues and -vectors of a symmetric matrix A."
    return eig(A, eig=np.linalg.eigh)


def eigvals(A, shear=False, eig=np.linalg.eig):
    "Eigenvalues (and optional principal shear values) of a matrix A."
    wA = eig(A.T)[0].T
    if shear:
        dim = wA.shape[0]
        if dim == 3:
            ij = [(1, 0), (2, 0), (2, 1)]
        elif dim == 2:
            ij = [(1, 0)]
        dwA = np.array([wA[i] - wA[j] for i, j in ij])
        return np.vstack((wA, dwA))
    else:
        return wA


def eigvalsh(A, shear=False):
    "Eigenvalues (and optional principal shear values) of a symmetric matrix A."
    return eigvals(A, shear=shear, eig=np.linalg.eigh)


def transpose(A, mode=1):
    "Transpose (mode=1) or major-transpose (mode=2) of matrix A."
    if mode == 1:
        return np.einsum("ij...->ji...", A)
    elif mode == 2:
        return np.einsum("ijkl...->klij...", A)
    else:
        raise ValueError("Unknown value of mode.")


def majortranspose(A):
    return transpose(A, mode=2)


def trace(A, out=None):
    r"""Return the sums along the diagonals of second order tensors.

    Parameters
    ----------
    A : ndarray of shape (M, M, ...)
        The array of second-order tensors.
    out : ndarray or None, optional
        If provided, the calculation is done into this array.

    Returns
    -------
    ndarray of shape (...)
        Array with the sums along the diagonals.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The trace of a second-order tensor is obtained by Eq. :eq:`math-trace`.

    ..  math::
        :label: math-trace

        \text{tr} \left( \boldsymbol{A} \right) &= \boldsymbol{A} : \boldsymbol{1}

        A_{ii} &= A_{ij} : \delta_{ij}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> A[..., 0]
    >>> A.shape
    (3, 3, 1)

    >>> A[..., 0]
    array([[0., 1., 2.],
           [3., 4., 5.],
           [6., 7., 8.]])

    >>> fem.math.trace(A)[..., 0]
    array(12.)

    See Also
    --------
    numpy.trace : Return the sum along diagonals of the array.

    """
    return np.trace(A, out=out)


def cdya_ik(A, B, parallel=False, **kwargs):
    r"""Return the ik-crossed dyadic product of two second-order tensors.

    Parameters
    ----------
    A : ndarray of shape ((N, N, ...)
        Array with second-order tensors.
    B : ndarray of shape (M, M, ...)
        Array with second-order tensors.
    parallel : bool, optional
        A flag to enable a threaded evaluation of the results (default is False).
    **kwargs : dict, optional
        Optional keyword-arguments for :func:`numpy.einsum`, e.g. ``out=None``.

    Returns
    -------
    ndarray of shape  (N, M, N, M, ...)
        The array of ik-crossed dyadic products.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions. The definition of the ik-crossed dyadic product is
    given in Eq. :eq:`math-cdya-ik`.

    ..  math::
        :label: math-cdya-ik

        \mathbb{C} &= \boldsymbol{A} \overset{ik}{\otimes} \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{ik}\ B_{jl}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> B = fem.math.transpose(np.arange(100, 109, dtype=float).reshape(1, 3, 3).T)
    >>> C = fem.math.cdya_ik(A, B)
    >>> C.shape
    (3, 3, 3, 3, 1)

    >>> C[..., 0].reshape(9, 9)
    array([[  0.,   0.,   0., 100., 101., 102., 200., 202., 204.],
           [  0.,   0.,   0., 103., 104., 105., 206., 208., 210.],
           [  0.,   0.,   0., 106., 107., 108., 212., 214., 216.],
           [300., 303., 306., 400., 404., 408., 500., 505., 510.],
           [309., 312., 315., 412., 416., 420., 515., 520., 525.],
           [318., 321., 324., 424., 428., 432., 530., 535., 540.],
           [600., 606., 612., 700., 707., 714., 800., 808., 816.],
           [618., 624., 630., 721., 728., 735., 824., 832., 840.],
           [636., 642., 648., 742., 749., 756., 848., 856., 864.]])

    See Also
    --------
    felupe.math.dya : Dyadic product of two first-order or two second-order tensors.
    felupe.math.cdya : Crossed dyadic product of two second-order tensors.
    felupe.math.cdya_il : il-crossed dyadic product of two second-order tensors.

    """
    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum
    return einsum("ij...,kl...->ikjl...", A, B, **kwargs)


def cdya_il(A, B, parallel=False, **kwargs):
    r"""Return the il-crossed dyadic product of two second-order tensors.

    Parameters
    ----------
    A : ndarray of shape ((N, N, ...)
        Array with second-order tensors.
    B : ndarray of shape (M, M, ...)
        Array with second-order tensors.
    parallel : bool, optional
        A flag to enable a threaded evaluation of the results (default is False).
    **kwargs : dict, optional
        Optional keyword-arguments for :func:`numpy.einsum`, e.g. ``out=None``.

    Returns
    -------
    ndarray of shape  (N, M, N, M, ...)
        The array of ik-crossed dyadic products.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions. The definition of the il-crossed dyadic product is
    given in Eq. :eq:`math-cdya-il`.

    ..  math::
        :label: math-cdya-il

        \mathbb{C} &= \boldsymbol{A} \overset{il}{\otimes} \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{il}\ B_{kj}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> B = fem.math.transpose(np.arange(100, 109, dtype=float).reshape(1, 3, 3).T)
    >>> C = fem.math.cdya_il(A, B)
    >>> C.shape
    (3, 3, 3, 3, 1)

    >>> C[..., 0].reshape(9, 9)
    array([[  0., 100., 200.,   0., 103., 206.,   0., 106., 212.],
           [  0., 101., 202.,   0., 104., 208.,   0., 107., 214.],
           [  0., 102., 204.,   0., 105., 210.,   0., 108., 216.],
           [300., 400., 500., 309., 412., 515., 318., 424., 530.],
           [303., 404., 505., 312., 416., 520., 321., 428., 535.],
           [306., 408., 510., 315., 420., 525., 324., 432., 540.],
           [600., 700., 800., 618., 721., 824., 636., 742., 848.],
           [606., 707., 808., 624., 728., 832., 642., 749., 856.],
           [612., 714., 816., 630., 735., 840., 648., 756., 864.]])

    See Also
    --------
    felupe.math.dya : Dyadic product of two first-order or two second-order tensors.
    felupe.math.cdya_ik : ik-crossed dyadic product of two second-order tensors.
    felupe.math.cdya : Crossed dyadic product of two second-order tensors.

    """
    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum
    return einsum("ij...,kl...->ilkj...", A, B, **kwargs)


def cdya(A, B, parallel=False, out=None, **kwargs):
    r"""Return the crossed dyadic product of two second-order tensors.

    Parameters
    ----------
    A : ndarray of shape ((M, M, ...)
        Array with second-order tensors.
    B : ndarray of shape (M, M, ...)
        Array with second-order tensors.
    parallel : bool, optional
        A flag to enable a threaded evaluation of the results (default is False).
    **kwargs : dict, optional
        Optional keyword-arguments for :func:`numpy.einsum`, e.g. ``out=None``.

    Returns
    -------
    ndarray of shape  (M, M, M, M, ...)
        The array of ik-crossed dyadic products.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions. The definition of the crossed dyadic product is
    given in Eq. :eq:`math-cdya`.

    ..  math::
        :label: math-cdya

        \mathbb{C} &= \boldsymbol{A} \odot \boldsymbol{B} = \frac{1}{2} \left(
            \boldsymbol{A} \overset{ik}{\otimes} \boldsymbol{B} +
            \boldsymbol{A} \overset{il}{\otimes} \boldsymbol{B}
        \right)

        \mathbb{C}_{ijkl} &= \frac{1}{2} \left( A_{ik}~B_{jl} + A_{il}~B_{kj} \right)

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> B = fem.math.transpose(np.arange(100, 109, dtype=float).reshape(1, 3, 3).T)
    >>> C = fem.math.cdya(A, B)
    >>> C.shape
    (3, 3, 3, 3, 1)

    >>> C[..., 0].reshape(9, 9)
    array([[  0. ,  50. , 100. ,  50. , 102. , 154. , 100. , 154. , 208. ],
           [  0. ,  50.5, 101. ,  51.5, 104. , 156.5, 103. , 157.5, 212. ],
           [  0. ,  51. , 102. ,  53. , 106. , 159. , 106. , 161. , 216. ],
           [300. , 351.5, 403. , 354.5, 408. , 461.5, 409. , 464.5, 520. ],
           [306. , 358. , 410. , 362. , 416. , 470. , 418. , 474. , 530. ],
           [312. , 364.5, 417. , 369.5, 424. , 478.5, 427. , 483.5, 540. ],
           [600. , 653. , 706. , 659. , 714. , 769. , 718. , 775. , 832. ],
           [612. , 665.5, 719. , 672.5, 728. , 783.5, 733. , 790.5, 848. ],
           [624. , 678. , 732. , 686. , 742. , 798. , 748. , 806. , 864. ]])

    See Also
    --------
    felupe.math.dya : Dyadic product of two first-order or two second-order tensors.
    felupe.math.cdya_ik : ik-crossed dyadic product of two second-order tensors.
    felupe.math.cdya_il : il-crossed dyadic product of two second-order tensors.

    """
    res = cdya_ik(A, B, parallel=parallel, out=out, **kwargs)
    res = np.add(res, cdya_il(A, B, parallel=parallel, **kwargs), out=res)
    return np.multiply(res, 0.5, out=res)


def cross(a, b):
    "Cross product of two vectors a and b."
    return np.einsum(
        "...i->i...", np.cross(np.einsum("i...->...i", a), np.einsum("i...->...i", b))
    )


def dot(A, B, mode=(2, 2), parallel=False, **kwargs):
    "Dot-product of A and B with inputs of n trailing axes.."

    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum

    if mode == (2, 2):
        return einsum("ik...,kj...->ij...", A, B, **kwargs)
    elif mode == (1, 1):
        return einsum("i...,i...->...", A, B, **kwargs)
    elif mode == (4, 4):
        return einsum("ijkp...,plmn...->ijklmn...", A, B, **kwargs)
    elif mode == (2, 1):
        return einsum("ij...,j...->i...", A, B, **kwargs)
    elif mode == (1, 2):
        return einsum("i...,ij...->j...", A, B, **kwargs)
    elif mode == (2, 3):
        return einsum("im...,mjk...->ijk...", A, B, **kwargs)
    elif mode == (3, 2):
        return einsum("ijm...,mk...->ijk...", A, B, **kwargs)
    elif mode == (4, 1):
        return einsum("ijkl...,l...->ijk...", A, B, **kwargs)
    elif mode == (1, 4):
        return einsum("i...,ijkl...->jkl...", A, B, **kwargs)
    elif mode == (2, 4):
        return einsum("im...,mjkl...->ijkl...", A, B, **kwargs)
    elif mode == (4, 2):
        return einsum("ijkm...,ml...->ijkl...", A, B, **kwargs)
    else:
        raise TypeError("Unknown shape of A and B.")


def ddot(A, B, mode=(2, 2), parallel=False, **kwargs):
    r"""Return the double-dot products of first-, second-, third- or fourth-order
    tensors.

    Parameters
    ----------
    A : ndarray of shape (M, M, ...)
        Array with first-, second-, third- or fourth-order tensors.
    B : ndarray of shape (M, M, ...)
        Array with first-, second-, third- or fourth-order tensors.
    mode : tuple of int, optional
        Mode of operation. Return the double-dot products of two second-order tensors
        with (2, 2), of a second-order and a fourth-order tensor with (2, 4) and so on.
        Default is (2, 2).
    parallel : bool, optional
        A flag to enable a threaded evaluation of the results (default is False).
    **kwargs : dict, optional
        Optional keyword-arguments for :func:`numpy.einsum`, e.g. ``out=None``.

    Returns
    -------
    ndarray of shape (...)
        Array with the double-dot products.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The double-dot product is obtained by Eq. :eq:`math-ddot22` for two second-order
    tensors,

    ..  math::
        :label: math-ddot22

        c &= \boldsymbol{A} : \boldsymbol{B}

        c &= A_{ij} : B_{ij}

    by Eq. :eq:`math-ddot44` for two fourth-order tensors,

    ..  math::
        :label: math-ddot44

        \mathbb{C} &= \mathbb{A} : \mathbb{B}

        \mathbb{C}_{ijmn} &=  \mathbb{A}_{ijkl} : \mathbb{B}_{klmn}

    by Eq. :eq:`math-ddot23` for a second-order and a third-order tensor,

    ..  math::
        :label: math-ddot23

        \boldsymbol{c} = \boldsymbol{A} : \mathcal{B}

        \qquad c_{k} &= A_{ij} : \mathcal{B}_{ijk}

    by Eq. :eq:`math-ddot32` for a third-order and a second-order tensor,

    ..  math::
        :label: math-ddot32

        \boldsymbol{c} = \mathcal{A} : \boldsymbol{A}

        \qquad c_{i} &= \mathcal{A}_{ijk} : B_{jk}

    by Eq. :eq:`math-ddot24` for a second-order and a fourth-order tensor

    ..  math::
        :label: math-ddot24

        \boldsymbol{C} &= \boldsymbol{A} : \mathbb{B}

        C_{kl} &= A_{ij} : \mathbb{B}_{ijkl}

    and by Eq. :eq:`math-ddot42` for a fourth-order and a second-order tensor.

    ..  math::
        :label: math-ddot42

        \boldsymbol{C} &= \mathbb{A} : \boldsymbol{A}

        C_{ij} &= \mathbb{A}_{ijkl} : B_{kl}


    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> A = fem.math.transpose(np.arange(9, dtype=float).reshape(1, 3, 3).T)
    >>> B = A + 10

    >>> fem.math.ddot(A, B)[..., 0]
    array(564.)

    >>> fem.math.ddot(fem.math.dya(A, A), B, mode=(4, 2))[..., 0]
    array([[   0.,  564., 1128.],
           [1692., 2256., 2820.],
           [3384., 3948., 4512.]])

    """

    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum

    if mode == (2, 2):
        return einsum("ij...,ij...->...", A, B, **kwargs)
    elif mode == (2, 4):
        return einsum("ij...,ijkl...->kl...", A, B, **kwargs)
    elif mode == (4, 2):
        return einsum("ijkl...,kl...->ij...", A, B, **kwargs)
    elif mode == (2, 3):
        return einsum("ij...,ijk...->k...", A, B, **kwargs)
    elif mode == (3, 2):
        return einsum("ijk...,jk...->i...", A, B, **kwargs)
    elif mode == (4, 4):
        return einsum("ijkl...,klmn...->ijmn...", A, B, **kwargs)
    else:
        raise TypeError("Unknown shape of A and B.")


def tovoigt(A, strain=False):
    r"""Return a three-dimensional second-order tensor in reduced symmetric (Voigt-
    notation) vector/matrix storage.

    Parameters
    ----------
    A : ndarray of shape (2, 2, ...) or (3, 3, ...)
        Array with three-dimensional second-order tensors.
    strain : bool, optional
        A flag to double the off-diagonal (shear) values for strain-like tensors.
        Default is False.

    Returns
    -------
    ndarray of shape (3, ...) or (6, ...)
        A three-dimensional second-order tensor in reduced symmetric (Voigt) vector/
        matrix storage.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.
    
    For a symmetric three-dimensional second-order tensor :math:`C_{ij} = C_{ji}`, the
    upper triangle entries are inserted into a 6x1 vector, starting from the main
    diagonal, followed by the consecutive next upper diagonals.

    ..  math::

        \boldsymbol{C} = \begin{bmatrix}
            C_{11} & C_{12} & C_{13} \\
            C_{12} & C_{22} & C_{23} \\
            C_{13} & C_{23} & C_{33}
        \end{bmatrix} \qquad \longrightarrow \boldsymbol{C} = \begin{bmatrix}
            C_{11} \\ C_{22} \\ C_{33} \\ C_{12} \\ C_{23} \\ C_{13}
        \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem
    >>> import numpy as np
    >>>
    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3, 1)
    >>> C[..., 0]
    array([[1. , 1.3, 1.5],
           [1.3, 1.1, 1.4],
           [1.5, 1.4, 1.2]])
    
    >>> fem.math.tovoigt(C)[..., 0]
    array([1. , 1.1, 1.2, 1.3, 1.4, 1.5])

    """
    dim = A.shape[:2]
    if dim == (2, 2):
        B = np.zeros((3, *A.shape[2:]))
        ij = [(0, 0), (1, 1), (0, 1)]
    elif dim == (3, 3):
        B = np.zeros((6, *A.shape[2:]))
        ij = [(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)]
    else:
        raise TypeError("Input shape must be (2, 2, ...) or (3, 3, ...).")
    for i6, (i, j) in enumerate(ij):
        B[i6] = A[i, j]
    if strain:
        B[dim[0] :] *= 2
    return B


def reshape(A, shape, trailing_axes=2):
    return A.reshape(np.append(shape, A.shape[-trailing_axes:]))


def ravel(A, trailing_axes=2):
    ij, shape = np.split(A.shape, [-trailing_axes])
    return reshape(A, shape=np.prod(ij))


def equivalent_von_mises(A):
    r"""Return the Equivalent von Mises values of symmetric second order-tensors.

    Parameters
    ----------
    A : ndarray of shape (3, 3, ...)
        Symmetric second-order tensors for which the equivalent von Mises values will be
        computed. The shape

    Returns
    -------
    ndarray of shape (...)
        The equivalent von Mises values.

    Notes
    -----
    The first two axes are the tensor dimensions and all remaining trailing axes are
    treated as batch dimensions.

    The equivalent von Mises value of a three-dimensional symmetric second order-tensor
    is given in Eq. :eq:`math-vonmises`.

    ..  math::
        :label: math-vonmises

        \boldsymbol{A}_{v} = \sqrt{
            \frac{3}{2} \text{dev}(\boldsymbol{A}) : \text{dev}(\boldsymbol{A})
        }

    Examples
    --------
    >>> import numpy as np
    >>> import felupe as fem
    >>>
    >>> stress = np.array([1, 1.1, 1.2, 1.1, 1.3, 1.4, 1.2, 1.4, 1.5]).reshape(3, 3, 1)
    >>> fem.math.equivalent_von_mises(stress)
    array([3.74432905])

    >>> stress = np.diag([3, 1, 1]).reshape(3, 3, 1)
    >>> fem.math.equivalent_von_mises(stress)
    array([2.])
    """

    if A.shape[:2] != (3, 3):
        raise TypeError("Only three-dimensional second-order tensors are supported.")

    devA = dev(A)
    return np.sqrt(3 / 2 * ddot(devA, devA))



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from .._integral import IntegralForm
from ._bilinear import BilinearForm
from ._linear import LinearForm


class LinearFormExpression:
    r"""A linear form object with methods for integration and assembly of
    vectors where ``v`` is a basis object on a field container.

    ..  math::

        L(v) = \int_\Omega f \cdot v \ dx

    Parameters
    ----------
    v : Basis
        An object with basis functions (gradients) of a field.
    grad_v : tuple of bool, optional (default is None)
        Flag to use the gradients of ``v``.


    """

    def __init__(self, v, grad_v=None):
        self.v = v
        self.dx = self.v.field[0].region.dV
        self._form = IntegralForm(
            np.zeros(len(v.field.fields)), self.v.field, self.dx, grad_v=grad_v
        )

        if grad_v is None:
            self.grad_v = np.zeros_like(self.v.field.fields, dtype=bool)
            self.grad_v[0] = True
        else:
            self.grad_v = grad_v

        self._linearform = [
            LinearForm(v=vi, grad_v=gvi, dx=self.dx)
            for vi, gvi in zip(self.v, self.grad_v)
        ]

    def integrate(self, weakform, args=(), kwargs={}, parallel=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kwargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) vector values.
        """

        return [
            form.integrate(fun, args, kwargs, parallel=parallel)
            for form, fun in zip(self._linearform, weakform)
        ]

    def assemble(self, weakform, args=(), kwargs={}, parallel=False):
        r"""Return the assembled integral as vector.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kwargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : csr_matrix
            The assembled vector.
        """

        values = self.integrate(weakform, args, kwargs, parallel=parallel)

        return self._form.assemble(values)


class BilinearFormExpression:
    r"""A bilinear form object with methods for integration and assembly of
    matrices where ``v`` is a mixed-basis object on mixed-fields.

    ..  math::

        a(v, u) = \int_\Omega v \cdot f \cdot u \ dx

    Parameters
    ----------
    v : Basis
        An object with basis functions (gradients) of a field.
    grad_v : tuple of bool, optional (default is None)
        Tuple of flags to use the gradients of ``v``.
    u : Basis
        An object with basis function (gradients) of a field.
    grad_u : tuple of bool, optional (default is None)
        Flag to use the gradients of ``u``.

    """

    def __init__(self, v, u, grad_v=None, grad_u=None):
        self.v = v
        self.u = u
        self.dx = self.v.field[0].region.dV

        self.nv = len(v.field.fields)
        self.i, self.j = np.triu_indices(self.nv)

        def _set_first_grad_true(grad, fields):
            if grad is None:
                grad = np.zeros_like(fields, dtype=bool)
                grad[0] = True
            return grad

        self.grad_v = _set_first_grad_true(grad_v, self.v.field.fields)
        self.grad_u = _set_first_grad_true(grad_u, self.u.field.fields)

        self._form = IntegralForm(
            fun=np.zeros(len(self.i)),
            v=self.v.field,
            dV=self.dx,
            u=self.u.field,
            grad_v=self.grad_v,
            grad_u=self.grad_u,
        )

        self._bilinearform = []

        for a, (i, j) in enumerate(zip(self.i, self.j)):
            self._bilinearform.append(
                BilinearForm(
                    v=self.v[i],
                    u=self.u[j],
                    grad_v=self.grad_v[i],
                    grad_u=self.grad_u[j],
                    dx=self.dx,
                )
            )

    def integrate(self, weakform, args=(), kwargs={}, parallel=False, sym=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kawargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) matrix values.
        """

        return [
            form.integrate(fun, args, kwargs, parallel=parallel, sym=sym)
            for form, fun in zip(self._bilinearform, weakform)
        ]

    def assemble(self, weakform, args=(), kwargs={}, parallel=False, sym=False):
        r"""Return the assembled integral as matrix.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kawargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : csr_matrix
            The assembled matrix.
        """

        values = self.integrate(weakform, args, kwargs, parallel=parallel, sym=sym)

        return self._form.assemble(values)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import tensortrax as tr

from ..math import cdya_ik, dot, transpose
from ._user_materials import Material


class Hyperelastic(Material):
    r"""A hyperelastic material definition with a given function for the strain energy
    density function per unit undeformed volume with Automatic Differentiation provided
    by ``tensortrax``.

    Parameters
    ----------
    fun : callable
        A strain energy density function in terms of the right Cauchy-Green deformation
        tensor :math:`\boldsymbol{C}`. Function signature must be
        ``fun = lambda C, **kwargs: psi`` for functions without state variables and
        ``fun = lambda C, statevars, **kwargs: [psi, statevars_new]`` for functions
        with state variables. The right Cauchy-Green deformation tensor will be a
        :class:`tensortrax.Tensor` when the function is evaluated. It is important to
        only use differentiable math-functions from ``tensortrax.math``!
    nstatevars : int, optional
        Number of state variables (default is 0).
    parallel : bool, optional
        A flag to invoke threaded strain energy density function evaluations (default
        is False). May introduce additional overhead for small-sized problems.
    **kwargs : dict, optional
        Optional keyword-arguments for the strain energy density function.

    Notes
    -----
    The strain energy density function :math:`\psi` must be given in terms of the right
    Cauchy-Green deformation tensor
    :math:`\boldsymbol{C} = \boldsymbol{F}^T \boldsymbol{F}`.

    ..  warning::
        It is important to only use differentiable math-functions from
        ``tensortrax.math``!

    Take this minimal code-block as template

    ..  math::

        \psi = \psi(\boldsymbol{C})

    ..  code-block::

        import tensortrax.math as tm

        def neo_hooke(C, mu):
            "Strain energy function of the Neo-Hookean material formulation."
            return mu / 2 * (tm.linalg.det(C) ** (-1/3) * tm.trace(C) - 3)

        umat = fem.Hyperelastic(neo_hooke, mu=1)

    and this code-block for material formulations with state variables.

    ..  math::

        \psi = \psi(\boldsymbol{C}, \boldsymbol{\zeta})

    ..  code-block::

        import tensortrax.math as tm

        def viscoelastic(C, Cin, mu, eta, dtime):
            "Finite strain viscoelastic material formulation."

            # unimodular part of the right Cauchy-Green deformation tensor
            Cu = tm.linalg.det(C) ** (-1 / 3) * C

            # update of state variables by evolution equation
            Ci = tm.special.from_triu_1d(Cin, like=C) + mu / eta * dtime * Cu
            Ci = tm.linalg.det(Ci) ** (-1 / 3) * Ci

            # first invariant of elastic part of right Cauchy-Green deformation tensor
            I1 = tm.trace(Cu @ tm.linalg.inv(Ci))

            # strain energy function and state variable
            return mu / 2 * (I1 - 3), tm.special.triu_1d(Ci)

        umat = fem.Hyperelastic(
            viscoelastic, mu=1, eta=1, dtime=1, nstatevars=6
        )

    ..  note::
        See the `documentation of tensortrax <https://github.com/adtzlr/tensortrax>`_
        for further details.

    Examples
    --------
    View force-stretch curves on elementary incompressible deformations.

    >>> import felupe as fem
    >>> import tensortrax.math as tm
    >>>
    >>> def neo_hooke(C, mu):
    >>>     "Strain energy function of the Neo-Hookean material formulation."
    >>>     return mu / 2 * (tm.linalg.det(C) ** (-1/3) * tm.trace(C) - 3)
    >>>
    >>> umat = fem.Hyperelastic(neo_hooke, mu=1)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_neo_hooke.png
        :width: 400px

    See Also
    --------
    saint_venant_kirchhoff : Strain energy function of the Saint
        Venant-Kirchhoff material formulation.
    neo_hooke : Strain energy function of the Neo-Hookean material formulation.
    mooney_rivlin : Strain energy function of the Mooney-Rivlin material formulation.
    yeoh : "Strain energy function of the Yeoh material formulation.
    third_order_deformation : Strain energy function of the
        Third-Order-Deformation material formulation.
    ogden : Strain energy function of the Ogden material formulation.
    arruda_boyce : Strain energy function of the Arruda-Boyce material formulation.
    extended_tube : Strain energy function of the Extended-Tube material formulation.
    van_der_waals : Strain energy function of the Van der Waals material formulation.
    finite_strain_viscoelastic : Finite strain viscoelastic material formulation.

    """

    def __init__(self, fun, nstatevars=0, parallel=False, **kwargs):
        if nstatevars > 0:
            # split the original function into two sub-functions
            self.fun = tr.take(fun, item=0)
            self.fun_statevars = tr.take(fun, item=1)
        else:
            self.fun = fun

        self.parallel = parallel

        super().__init__(
            stress=self._stress,
            elasticity=self._elasticity,
            nstatevars=nstatevars,
            **kwargs,
        )

    def _stress(self, x, **kwargs):
        F = np.ascontiguousarray(x[0])

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        C = dot(transpose(F), F)
        dWdC = tr.gradient(self.fun, wrt=0, ntrax=2, parallel=self.parallel, sym=True)(
            C, *statevars, **kwargs
        )
        if self.nstatevars > 0:
            statevars_new = tr.function(
                self.fun_statevars, wrt=0, ntrax=2, parallel=self.parallel
            )(C, *statevars, **kwargs)
        else:
            statevars_new = None
        return [dot(F, 2 * dWdC), statevars_new]

    def _elasticity(self, x, **kwargs):
        F = np.ascontiguousarray(x[0])

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        C = dot(transpose(F), F)
        d2WdCdC, dWdC, W = tr.hessian(
            self.fun, wrt=0, ntrax=2, full_output=True, parallel=self.parallel, sym=True
        )(C, *statevars, **kwargs)
        A = 4 * np.einsum(
            "iI...,kK...,IJKL...->iJkL...", F, F, np.ascontiguousarray(d2WdCdC)
        )
        B = cdya_ik(np.eye(3), 2 * dWdC)
        return [np.sum(np.broadcast_arrays(A, B), axis=0)]


class MaterialAD(Material):
    r"""A user-defined material definition with a given function for the partial
    derivative of the strain energy function w.r.t. the deformation gradient tensor
    with Automatic Differentiation provided by
    `tensortrax <https://github.com/adtzlr/tensortrax>`_.

    Parameters
    ----------
    fun : callable
        A gradient of the strain energy density function w.r.t. the deformation gradient
        tensor :math:`\boldsymbol{F}`. Function signature must be
        ``fun = lambda F, **kwargs: P`` for functions without state variables and
        ``fun = lambda F, statevars, **kwargs: [P, statevars_new]`` for functions
        with state variables. The deformation gradient tensor will be a
        :class:`tensortrax.Tensor` when the function is evaluated. It is important to
        only use differentiable math-functions from ``tensortrax.math``!
    nstatevars : int, optional
        Number of state variables (default is 0).
    parallel : bool, optional
        A flag to invoke threaded gradient of strain energy density function evaluations
        (default is False). May introduce additional overhead for small-sized problems.
    **kwargs : dict, optional
        Optional keyword-arguments for the gradient of the strain energy density
        function.

    Notes
    -----
    The gradient of the strain energy density function
    :math:`\frac{\partial \psi}{\partial \boldsymbol{F}}` must be given in terms of the
    deformation gradient tensor :math:`\boldsymbol{F}`.

    ..  warning::
        It is important to only use differentiable math-functions from
        ``tensortrax.math``!

    ..  math::

        \boldsymbol{P} = \frac{\partial \psi(\boldsymbol{F}, \boldsymbol{\zeta})}{
            \partial \boldsymbol{F}}

    Take this code-block as template

    ..  code-block::

        import tensortrax.math as tm

        def neo_hooke(F, mu):
            "First Piola-Kirchhoff stress of the Neo-Hookean material formulation."

            C = tm.dot(tm.transpose(F), F)
            Cu = tm.linalg.det(C) ** (-1/3) * C

            return mu * F @ tm.special.dev(Cu) @ tm.linalg.inv(C)

        umat = fem.MaterialAD(neo_hooke, mu=1)

    and this code-block for material formulations with state variables:

    ..  code-block::

        import tensortrax.math as tm

        def viscoelastic(F, Cin, mu, eta, dtime):
            "Finite strain viscoelastic material formulation."

            # unimodular part of the right Cauchy-Green deformation tensor
            C = tm.dot(tm.transpose(F), F)
            Cu = tm.linalg.det(C) ** (-1 / 3) * C

            # update of state variables by evolution equation
            Ci = tm.special.from_triu_1d(Cin, like=C) + mu / eta * dtime * Cu
            Ci = tm.linalg.det(Ci) ** (-1 / 3) * Ci

            # second Piola-Kirchhoff stress tensor
            S = mu * tm.special.dev(Cu @ tm.linalg.inv(Ci)) @ tm.linalg.inv(C)

            # first Piola-Kirchhoff stress tensor and state variable
            return F @ S, tm.special.triu_1d(Ci)

        umat = fem.MaterialAD(
            viscoelastic, mu=1, eta=1, dtime=1, nstatevars=6
        )

    ..  note::
        See the `documentation of tensortrax <https://github.com/adtzlr/tensortrax>`_
        for further details.

    Examples
    --------
    >>> import felupe as fem
    >>> import tensortrax.math as tm
    >>>
    >>> def neo_hooke(F, mu):
    >>>     C = tm.dot(tm.transpose(F), F)
    >>>     S = mu * tm.special.dev(tm.linalg.det(C)**(-1/3) * C) @ tm.linalg.inv(C)
    >>>     return F @ S
    >>>
    >>> umat = fem.MaterialAD(neo_hooke, mu=1)
    >>> ax = umat.plot(incompressible=True)

    ..  image:: images/umat_materialad_neohooke.png
        :width: 400px

    """

    def __init__(self, fun, nstatevars=0, parallel=False, **kwargs):
        if nstatevars > 0:
            # split the original function into two sub-functions
            self.fun = tr.take(fun, item=0)
            self.fun_statevars = tr.take(fun, item=1)
        else:
            self.fun = fun

        self.parallel = parallel

        super().__init__(
            stress=self._stress,
            elasticity=self._elasticity,
            nstatevars=nstatevars,
            **kwargs,
        )

    def _stress(self, x, **kwargs):
        F = np.ascontiguousarray(x[0])

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        dWdF = tr.function(self.fun, wrt=0, ntrax=2, parallel=self.parallel)(
            F, *statevars, **kwargs
        )
        if self.nstatevars > 0:
            statevars_new = tr.function(
                self.fun_statevars, wrt=0, ntrax=2, parallel=self.parallel
            )(F, *statevars, **kwargs)
        else:
            statevars_new = None
        return [dWdF, statevars_new]

    def _elasticity(self, x, **kwargs):
        F = np.ascontiguousarray(x[0])

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        d2WdFdF = tr.jacobian(self.fun, wrt=0, ntrax=2, parallel=self.parallel)(
            F, *statevars, **kwargs
        )
        return [d2WdFdF]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def linsteps(points, num=10, endpoint=True, axis=None, axes=None):
    """Return a sequence from batches of evenly spaced samples between pairs of
    milestone points.

    Parameters
    ----------
    points : array_like
        The milestone values of the sequence.
    num : int, optional
        Number of samples (without the end point) to generate. Must be non-negative
        (default is 10).
    endpoint : bool, optional
        If True, ``points[-1]`` is the last sample. Otherwise, it is not included
        (default is True).
    axis : int or None, optional
        The axis in the result to store the samples. By default (None), the samples will
        be concatenated along the existing first axis. If ``axis > 0``, the samples will
        be inserted at column ``axis``. Only positive integers are supported.
    axes : int or None, optional
        Number of output columns if ``axis`` is not None. Requires ``axes > axis`` and
        will be set to ``axes = axis + 1`` by default (None).

    Returns
    -------
    samples : ndarray
        Concatenated batches of ``num`` equally spaced samples.

    Examples
    --------
    >>> import felupe as fem

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=2)
    array([0.  , 0.25, 0.5 , 1.  , 1.5 , 2.5 , 3.5 ])

    Not including the end value of the sequence does not change the step size.

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=2, endpoint=False)
    array([0.  , 0.25, 0.5 , 1.  , 1.5 , 2.5 ])

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=2, axis=1)
    array([[0.  , 0.  ],
           [0.  , 0.25],
           [0.  , 0.5 ],
           [0.  , 1.  ],
           [0.  , 1.5 ],
           [0.  , 2.5 ],
           [0.  , 3.5 ]])

    Output with four columns:

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=2, axis=1, axes=4)
    array([[0.  , 0.  , 0.  , 0.  ],
           [0.  , 0.25, 0.  , 0.  ],
           [0.  , 0.5 , 0.  , 0.  ],
           [0.  , 1.  , 0.  , 0.  ],
           [0.  , 1.5 , 0.  , 0.  ],
           [0.  , 2.5 , 0.  , 0.  ],
           [0.  , 3.5 , 0.  , 0.  ]])

    The ouput degenerates to the end value

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=0)
    array([3.5])

    or an empty array.

    >>> fem.math.linsteps([0, 0.5, 1.5, 3.5], num=0, endpoint=False)
    array([], dtype=float64)

    See Also
    --------
    numpy.linspace : Return evenly spaced numbers over a specified interval.
    """
    points = np.array(points).ravel()
    start = points[:-1]
    end = points[1:]
    num = np.array([num]).ravel()

    if len(num) == 1:
        num = np.tile(num, max(1, len(start)))

    num = np.pad(num, (0, max(0, len(start) - len(num))), mode="edge")

    steplist = [
        np.linspace(a, b, n, endpoint=False) for a, b, n in zip(start, end, num)
    ]
    if len(steplist) > 0:
        steps = np.concatenate(
            [np.linspace(a, b, n, endpoint=False) for a, b, n in zip(start, end, num)]
        )
    else:
        steps = np.array([])

    if endpoint:
        steps = np.append(steps, points[-1])

    if axis is not None:
        if axes is None:
            axes = axis + 1

        steps_1d = steps
        steps = np.zeros((len(steps_1d), axes))
        steps[:, axis] = steps_1d

    return steps



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..math import cdya_ik, ddot, dya, identity, sqrt, trace


def linear_elastic(dε, εn, σn, ζn, λ, μ, **kwargs):
    r"""3D linear-elastic material formulation to be used in
    :class:`~felupe.MaterialStrain`.

    Arguments
    ---------
    dε : ndarray
        Strain increment.
    εn : ndarray
        Old strain tensor.
    σn : ndarray
        Old stress tensor.
    ζn : list
        List of old state variables.
    λ : float
        First Lamé-constant.
    μ : float
        Second Lamé-constant (shear modulus).

    Returns
    -------
    dσdε : ndarray
        Elasticity tensor.
    σ : ndarray
        (New) stress tensor.
    ζ : list
        List of new state variables.

    Notes
    -----

    1.  Given state in point :math:`\boldsymbol{x} (\boldsymbol{\sigma}_n)` (valid).

    2.  Given strain increment :math:`\Delta\boldsymbol{\varepsilon}`, so that
        :math:`\boldsymbol{\varepsilon} = \boldsymbol{\varepsilon}_n + \Delta\boldsymbol{\varepsilon}`.

    3.  Evaluation of the stress :math:`\boldsymbol{\sigma}` and the algorithmic
        consistent tangent modulus :math:`\mathbb{C}` (=``dσdε``).

        ..  math::

            \mathbb{C} &= \lambda \ \boldsymbol{1} \otimes \boldsymbol{1} +
                2 \mu \ \boldsymbol{1} \odot \boldsymbol{1}

            \boldsymbol{\sigma} &= \boldsymbol{\sigma}_n
                + \mathbb{C} : \Delta\boldsymbol{\varepsilon}

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.MaterialStrain(material=fem.linear_elastic, λ=2.0, μ=1.0)

    See Also
    --------
    MaterialStrain : A strain-based user-defined material definition with a given
        function for the stress tensor and the (fourth-order) elasticity tensor.

    """

    # change of stress due to change of strain
    eye = identity(dim=3, shape=(1, 1))
    dσ = 2 * μ * dε + λ * trace(dε) * eye

    # update stress
    σ = σn + dσ

    # evaluate elasticity tensor
    if kwargs["tangent"]:
        dσdε = 2 * μ * cdya_ik(eye, eye) + λ * dya(eye, eye)
    else:
        dσdε = None

    # update state variables (not used here)
    ζ = ζn

    return dσdε, σ, ζ


def linear_elastic_plastic_isotropic_hardening(dε, εn, σn, ζn, λ, μ, σy, K, **kwargs):
    r"""Linear-elastic-plastic material formulation with linear isotropic
    hardening (return mapping algorithm) to be used in :class:`~felupe.MaterialStrain`.

    Arguments
    ---------
    dε : ndarray
        Strain increment.
    εn : ndarray
        Old strain tensor.
    σn : ndarray
        Old stress tensor.
    ζn : list
        List of old state variables.
    λ : float
        First Lamé-constant.
    μ : float
        Second Lamé-constant (shear modulus).
    σy : float
        Initial yield stress.
    K : float
        Isotropic hardening modulus.

    Returns
    -------
    dσdε : ndarray
        Algorithmic consistent elasticity tensor.
    σ : ndarray
        (New) stress tensor.
    ζ : list
        List of new state variables.

    Notes
    -----

    1.  Given state in point :math:`x (\sigma_n, \zeta_n=[\varepsilon^p_n, \alpha_n])`
        (valid).

    2.  Given strain increment :math:`\Delta\varepsilon`, so that
        :math:`\varepsilon = \varepsilon_n + \Delta\varepsilon`.

    3.  Evaluation of the hypothetic trial state:

        ..  math::

            \mathbb{C} &= \lambda\ \boldsymbol{1} \otimes \boldsymbol{1}
                + 2 \mu\ \boldsymbol{1} \odot \boldsymbol{1}

            \sigma &= \sigma_n + \mathbb{C} : \Delta\varepsilon

            s &= \text{dev}(\sigma)

            \varepsilon^p &= \varepsilon^p_n

            \alpha &= \alpha_n

            f &= ||s|| - \sqrt{\frac{2}{3}}\ (\sigma_y + K \alpha)

    4.  If :math:`f \le 0`, then elastic step:

        Set :math:`y = y_n + \Delta y, y=(\sigma, \zeta=[\varepsilon^p, \alpha])`,

        algorithmic consistent tangent modulus :math:`d\sigma d\varepsilon`.

        ..  math::

            d\sigma d\varepsilon = \mathbb{C}

        Else:

        ..  math::

            d\gamma &= \frac{f}{2\mu + \frac{2}{3} K}

            n &= \frac{s}{||s||}

            \sigma &= \sigma - 2\mu \Delta\gamma n

            \varepsilon^p &= \varepsilon^p_n + \Delta\gamma n

            \alpha &= \alpha_n + \sqrt{\frac{2}{3}}\ \Delta\gamma

        Algorithmic consistent tangent modulus:

        ..  math::

            d\sigma d\varepsilon = \mathbb{C}
                - \frac{2 \mu}{1 + \frac{K}{3 \mu}} n \otimes n
                - \frac{2 \mu \Delta\gamma}{||s||} \left[
                    2 \mu \left( \boldsymbol{1} \odot \boldsymbol{1}
                     - \frac{1}{3} \boldsymbol{1} \otimes \boldsymbol{1}
                    - n \otimes n \right)
                \right]

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.MaterialStrain(
    >>>     material=fem.linear_elastic_plastic_isotropic_hardening,
    >>>     λ=2.0,
    >>>     μ=1.0,
    >>>     σy=1.0,
    >>>     K=0.1,
    >>>     dim=3,
    >>>     statevars=(1, (3, 3)),
    >>> )

    See Also
    --------
    MaterialStrain : A strain-based user-defined material definition with a given
        function for the stress tensor and the (fourth-order) elasticity tensor.

    """

    eye = identity(dε)

    # elasticity tensor
    if kwargs["tangent"]:
        dσdε = np.zeros((3, 3, 3, 3, *dε.shape[2:]))
        dσdε[:] = λ * dya(eye, eye) + 2 * μ * cdya_ik(eye, eye)
    else:
        dσdε = None

    # elastic hypothetic (trial) stress and deviatoric stress
    dσ = 2 * μ * dε + λ * trace(dε) * eye
    σ = σn + dσ
    s = σ - 1 / 3 * trace(σ) * eye

    # unpack old state variables
    α, εp = ζn

    # hypothetic (trial) yield function
    norm_s = sqrt(ddot(s, s))
    f = norm_s - sqrt(2 / 3) * (σy + K * α)

    ζ = ζn

    # check yield function and create a mask where plasticity occurs
    mask = (f > 0)[0]

    # update stress, tangent and state due to plasticity
    if np.any(mask):
        dγ = f / (2 * μ + 2 / 3 * K)
        n = s / norm_s
        εp = εp + dγ * n
        α = α + sqrt(2 / 3) * dγ

        # stress
        σ[..., mask] = (σ - 2 * μ * dγ * n)[..., mask]

        # algorithmic consistent tangent modulus
        if kwargs["tangent"]:
            dσdε[..., mask] = (
                dσdε
                - 2 * μ / (1 + K / (3 * μ)) * dya(n, n)
                - 2
                * μ
                * dγ
                / norm_s
                * (
                    2 * μ * (cdya_ik(eye, eye) - 1 / 3 * dya(eye, eye))
                    - 2 * μ * dya(n, n)
                )
            )[..., mask]

        # update list of state variables
        ζ[0][..., mask] = α[..., mask]
        ζ[1][..., mask] = εp[..., mask]

    return dσdε, σ, ζ



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix

from ..assembly import IntegralForm
from ._helpers import Assemble, Results


class SolidBodyGravity:
    r"""A gravity (body) force on a solid body.

    Parameters
    ----------
    field : FieldContainer
        A field container with fields created on a boundary region.
    gravity : ndarray or None, optional
        The prescribed values of gravity :math:`\boldsymbol{g}` (default is None). If
        None, the gravity vector is set to zero (the dimension of the gravity vector is
        derived from the first field of the field container).
    density : float, optional
        The density :math:`\rho` of the solid body (default is 1.0).

    Notes
    -----
    ..  math::

        \delta W_{ext} = \int_V
            \delta \boldsymbol{u} \cdot \rho \boldsymbol{g} \ dV

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Cube(n=6)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries = fem.dof.symmetry(field[0])
    >>>
    >>> umat = fem.NeoHooke(mu=1, bulk=2)
    >>> solid = fem.SolidBody(umat, field)
    >>> gravity = fem.SolidBodyGravity(field, density=1.0)
    >>>
    >>> table = fem.math.linsteps([0, 1], num=5, axis=0, axes=3)
    >>> step = fem.Step(
    >>>     items=[solid, gravity],
    >>>     ramp={gravity: 2 * table},
    >>>     boundaries=boundaries,
    >>> )
    >>>
    >>> job = fem.Job(steps=[step]).evaluate()
    >>> ax = solid.imshow("Principal Values of Cauchy Stress")

    """

    def __init__(self, field, gravity=None, density=1.0):
        self.field = field
        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)
        self._form = IntegralForm

        self.results.gravity = np.zeros(self.field[0].dim)
        if gravity is not None:
            self.results.gravity = np.array(gravity)

        self.results.density = density

    def update(self, gravity):
        self.__init__(self.field, gravity, self.results.density)

    def _vector(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        # copy and take only the first (displacement) field of the container
        f = self.field.copy()
        f.fields = f.fields[0:1]

        self.results.force = self._form(
            fun=[self.results.density * self.results.gravity.reshape(-1, 1, 1)],
            v=f,
            dV=self.field.region.dV,
            grad_v=[False],
        ).assemble(parallel=parallel)

        if len(self.field) > 1:
            self.results.force.resize(np.sum(self.field.fieldsizes), 1)

        return -self.results.force

    def _matrix(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        n = np.sum(self.field.fieldsizes)
        self.results.stiffness = csr_matrix(([0], ([0], [0])), shape=(n, n))

        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import inspect
import os
from time import perf_counter

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import spsolve

from .. import solve as fesolve
from ..assembly import IntegralForm
from ..math import norm


class NewtonResult:
    r"""A data class which represents the result found by Newton's method. All
    parameters are available as attributes.

    Parameters
    ----------
    x : felupe.FieldContainer or ndarray
        Array or Field container with values at a solution found by Newton's method.
    fun : ndarray or None, optional
        Values of objective function (default is None).
    jac : ndarray or None, optional
        Values of the Jacobian of the objective function (default is None).
    success : bool or None, optional
        A boolean flag which is True if the solution converged (default is None).
    iterations : int or None, optional
        Number of iterations until solution converged (default is None).
    xnorms : array of float or None, optional
        List with norms of the values of the solution (default is None).
    fnorms : float or None, optional
        List with norms of the objective function (default is None).

    Notes
    -----
    The objective function's norm is relative based on the function values on the
    prescribed degrees of freedom :math:`(\bullet)_0`. A small number
    :math:`\varepsilon` is added to avoid numeric instabilities.

    ..  math::

        \text{norm}(\boldsymbol{f}) = \frac{||\boldsymbol{f}_1||}
            {\varepsilon + ||\boldsymbol{f}_0||}

    """

    def __init__(
        self,
        x,
        fun=None,
        jac=None,
        success=None,
        iterations=None,
        xnorms=None,
        fnorms=None,
    ):
        self.x = x
        self.fun = fun
        self.jac = jac
        self.success = success
        self.iterations = iterations
        self.xnorms = xnorms
        self.fnorms = fnorms


def fun_items(items, x, parallel=False):
    "Assemble the sparse system vector for each item."

    # init keyword arguments
    kwargs = {"parallel": parallel}

    # link field of items with global field
    [item.field.link(x) for item in items]

    # init vector with shape from global field
    shape = (np.sum(x.fieldsizes), 1)
    vector = csr_matrix(shape)

    for body in items:
        # assemble vector
        r = body.assemble.vector(field=body.field, **kwargs)

        # check and reshape vector
        if r.shape != shape:
            r.resize(*shape)

        # add vector
        vector += r

    return vector.toarray()[:, 0]


def jac_items(items, x, parallel=False):
    "Assemble the sparse system matrix for each item."

    # init keyword arguments
    kwargs = {"parallel": parallel}

    # init matrix with shape from global field
    shape = (np.sum(x.fieldsizes), np.sum(x.fieldsizes))
    matrix = csr_matrix(shape)

    for body in items:
        # assemble matrix
        K = body.assemble.matrix(**kwargs)

        # check and reshape matrix
        if K.shape != matrix.shape:
            K.resize(*shape)

        # add matrix
        matrix += K

    return matrix


def fun(x, umat, parallel=False, grad=True, add_identity=True, sym=False):
    "Force residuals from assembly of equilibrium (weak form)."

    return (
        IntegralForm(
            fun=umat.gradient(x.extract(grad=grad, add_identity=add_identity, sym=sym))[
                :-1
            ],
            v=x,
            dV=x.region.dV,
        )
        .assemble(parallel=parallel)
        .toarray()[:, 0]
    )


def jac(x, umat, parallel=False, grad=True, add_identity=True, sym=False):
    "Tangent stiffness matrix from assembly of linearized equilibrium."

    return IntegralForm(
        fun=umat.hessian(x.extract(grad=grad, add_identity=add_identity, sym=sym)),
        v=x,
        dV=x.region.dV,
        u=x,
    ).assemble(parallel=parallel)


def solve(A, b, x, dof1, dof0, offsets=None, ext0=None, solver=spsolve):
    "Solve partitioned system."

    system = fesolve.partition(x, A, dof1, dof0, -b)
    dx = fesolve.solve(*system, ext0, solver=solver)

    return dx


def check(dx, x, f, xtol, ftol, dof1=None, dof0=None, items=None, eps=1e-3):
    "Check result."

    def sumnorm(x):
        return np.sum(norm(x))

    xnorm = sumnorm(dx)

    if dof1 is None:
        dof1 = slice(None)

    if dof0 is None:
        dof0 = slice(0, 0)

    fnorm = sumnorm(f[dof1]) / (eps + sumnorm(f[dof0]))
    success = fnorm < ftol and xnorm < xtol

    if success and items is not None:
        for item in items:
            [item.results.update_statevars() for item in items]

    return xnorm, fnorm, success


def update(x, dx):
    "Update field."
    # x += dx # in-place
    return x + dx


def newtonrhapson(
    x0=None,
    fun=fun,
    jac=jac,
    solve=solve,
    maxiter=16,
    update=update,
    check=check,
    args=(),
    kwargs={},
    tol=np.sqrt(np.finfo(float).eps),
    items=None,
    dof1=None,
    dof0=None,
    ext0=None,
    solver=spsolve,
    verbose=True,
):
    r"""Find a root of a real function using the Newton-Raphson method.

    Parameters
    ----------
    x0 : felupe.FieldContainer, ndarray or None (optional)
        Array or Field container with values of unknowns at a valid starting point
        (default is None).
    fun : callable, optional
        Callable which assembles the vector-valued objective function. Additional args
        and kwargs are passed. Function signature of a user-defined function has to be
        ``fun = lambda x, *args, **kwargs: f``.
    jac : callable, optional
        Callable which assembles the matrix-valued Jacobian. Additional args and kwargs
        are passed. Function signature of a user-defined Jacobian has to be
        ``jac = lambda x, *args, **kwargs: K``.
    solve : callable, optional
        Callable which prepares the linear equation system and solves it. If a keyword-
        argument from the list ``["x", "dof1", "dof0", "ext0", "solver"]`` is found in
        the function-signature, then these arguments are passed to ``solve``.
    maxiter : int, optional
        Maximum number of function iterations (default is 16).
    update : callable, optional
        Callable to update the unknowns. Function signature must be
        ``update = lambda x0, dx: x``.
    check : callable, optional
        Callable to the check the result.
    tol : float, optional
        Tolerance value to check if the function has converged (default is 1.490e-8).
    items : list or None, optional
        List with items which provide methods for assembly, e.g. like
        :class:`felupe.SolidBody` (default is None).
    dof1 : ndarray or None, optional
        1d-array of int with all active degrees of freedom (default is None).
    dof0 : ndarray or None, optional
        1d-array of int with all prescribed degress of freedom (default is None).
    ext0 : ndarray or None, optional
        Field values at mesh-points for the prescribed components of the unknowns based
        on ``dof0`` (default is None).
    solver : callable, optional
        A sparse or dense solver (default is :func:`scipy.sparse.linalg.spsolve`). For a
        more performant alternative install PyPardiso and use :func:`pypardiso.spsolve`.
    verbose : bool or int, optional
        Verbosity level: False or 0 for no logging, True or 1 for a progress bar and
        2 for a text-based logging output (default is True).If the environmental
        variable FELUPE_VERBOSE is set and its value is ``false``, then this argument is
        ignored and logging is turned off.

    Returns
    -------
    felupe.tools.NewtonResult
        The result object.

    Notes
    -----
    Nonlinear equilibrium equations :math:`f(x)` as a function of the unknowns :math:`x`
    are solved by linearization of :math:`f` at a valid starting point of given unknowns
    :math:`x_0`.

    ..  math::

        f(x_0) = 0

    The linearization is given by

    ..  math::

        f(x_0 + dx) \approx f(x_0) + K(x_0) \ dx \ (= 0)

    with the Jacobian, evaluated at given unknowns :math:`x_0`,

    ..  math::

        K(x_0) = \frac{\partial f}{\partial x}(x_0)

    and is rearranged to an equation system with left- and right-hand sides.

    ..  math::

        K(x_0) \ dx = -f(x_0)

    After a solution is found, the unknowns are updated.

    ..  math::

        dx &= \text{solve} \left( K(x_0), -f(x_0) \right)

        x &= x_0 + dx

    Repeated evaluations lead to an incrementally updated solution of :math:`x`. Herein
    :math:`x_n` refer to the inital unknowns whereas :math:`x` are the updated unknowns
    (the subscript :math:`(\bullet)_{n+1}` is dropped for readability).

    ..  math::
        :name: eq:newton-solve

        dx &= \text{solve} \left( K(x_n), -f(x_n) \right)

         x &= x_n + dx

    Then, the nonlinear equilibrium equations are evaluated with the updated unknowns
    :math:`f(x)`. The procedure is repeated until convergence is reached.

    Examples
    --------
    >>> import felupe as fem

    >>> region = fem.RegionHexahedron(fem.Cube(n=6))
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries, loadcase = fem.dof.uniaxial(field, move=0.2, clamped=True)
    >>> solid = fem.SolidBody(umat=fem.NeoHooke(mu=1.0, bulk=2.0), field=field)
    >>> res = fem.newtonrhapson(items=[solid], **loadcase)

    Newton's method had success

    >>> res.success
    True

    and 4 iterations were needed to converge within the specified tolerance.

    >>> res.iterations
    4

    The norm of the objective function for all active degrees of freedom is lower than
    3e-15.

    >>> np.linalg.norm(res.fun[loadcase["dof1"]])
    2.7482611016095555e-15

    """
    VERBOSE = os.environ.get("FELUPE_VERBOSE")
    if VERBOSE is None:
        verbose = verbose
    else:
        verbose = VERBOSE == "true"

    if verbose:
        runtimes = [perf_counter()]
        soltimes = []

    if x0 is not None:
        x = x0

    else:
        x = items[0].field

    kwargs_solve = {}
    sig = inspect.signature(solve)

    if items is not None:
        f = fun_items(items, x, *args, **kwargs)
    else:
        f = fun(x, *args, **kwargs)

    if verbose:
        print()
        print("Newton-Rhapson solver")
        print("=====================")
        print()
        print("| # | norm(fun) |  norm(dx) |")
        print("|---|-----------|-----------|")

    xnorms, fnorms = [], []

    # iteration loop
    for iteration in range(maxiter):
        if items is not None:
            K = jac_items(items, x, *args, **kwargs)
        else:
            K = jac(x, *args, **kwargs)

        # create keyword-arguments for solving the linear system
        keys = ["x", "dof1", "dof0", "ext0", "solver"]
        values = [x, dof1, dof0, ext0, solver]

        for key, value in zip(keys, values):
            if key in sig.parameters:
                kwargs_solve[key] = value

        if verbose:
            soltime_start = perf_counter()

        dx = solve(K, -f, **kwargs_solve)

        if verbose:
            soltime_end = perf_counter()
            soltimes.append([soltime_start, soltime_end])

        x = update(x, dx)

        if items is not None:
            f = fun_items(items, x, *args, **kwargs)
        else:
            f = fun(x, *args, **kwargs)

        xnorm, fnorm, success = check(
            dx=dx, x=x, f=f, xtol=np.inf, ftol=tol, dof1=dof1, dof0=dof0, items=items
        )
        xnorms.append(xnorm)
        fnorms.append(fnorm)

        if verbose:
            print("|%2d | %1.3e | %1.3e |" % (1 + iteration, fnorm, xnorm))

        if success:
            break

        if np.any(np.isnan([xnorm, fnorm])):
            raise ValueError("Norm of unknowns is NaN.")

    if 1 + iteration == maxiter and not success:
        raise ValueError("Maximum number of iterations reached (not converged).\n")

    Res = NewtonResult(
        x=x,
        fun=f,
        success=success,
        iterations=1 + iteration,
        xnorms=xnorms,
        fnorms=fnorms,
    )

    if verbose:
        runtimes.append(perf_counter())
        runtime = np.diff(runtimes)[0]
        soltime = np.diff(soltimes).sum()
        print(
            "\nConverged in %d iterations (Assembly: %1.4g s, Solve: %1.4g s).\n"
            % (iteration + 1, runtime - soltime, soltime)
        )

    return Res



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect

import numpy as np

from ..assembly import IntegralForm
from ..constitution import AreaChange
from ..math import det, dot, transpose
from ..tools._plot import ViewSolid
from ._helpers import Assemble, Evaluate, Results


class Solid:
    "Base class for solid bodies which provides methods for visualisations."

    def view(self, point_data=None, cell_data=None, cell_type=None):
        """View the solid with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).

        Returns
        -------
        felupe.ViewSolid
            An object which provides visualization methods for solid bodies,
            e.g. :class:`felupe.SolidBody`.

        See Also
        --------
        felupe.ViewSolid : Visualization methods for :class:`felupe.SolidBody`.
        """

        return ViewSolid(
            self.field,
            solid=self,
            point_data=point_data,
            cell_data=cell_data,
            cell_type=cell_type,
        )

    def plot(self, *args, **kwargs):
        """Plot the solid body.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        """
        return self.view().plot(*args, **kwargs)

    def screenshot(
        self,
        *args,
        filename="solidbody.png",
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the solid body.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, **kwargs):
        """Take a screenshot of the solid body, show the image data in a figure and
        return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax


class SolidBody(Solid):
    r"""A SolidBody with methods for the assembly of sparse vectors/matrices.

    Parameters
    ----------
    umat : class
        A class which provides methods for evaluating the gradient and the hessian of
        the strain energy density function per unit undeformed volume. The function
        signatures must be ``dψdF, ζ_new = umat.gradient([F, ζ])`` for
        the gradient and ``d2ψdFdF = umat.hessian([F, ζ])`` for the hessian of
        the strain energy density function :math:`\psi(\boldsymbol{F})`, where
        :math:`\boldsymbol{F}` is the deformation gradient tensor and :math:`\zeta`
        holds the array of internal state variables.
    field : FieldContainer
        A field container with one or more fields.
    statevars : ndarray or None, optional
        Array of initial internal state variables (default is None).

    Notes
    -----
    The total potential energy of internal forces is given in Eq.
    :eq:`solidbody`

    ..  math::
        :label: solidbody

        \Pi_{int}(\boldsymbol{F}) = \int_V \psi(\boldsymbol{F})\ dV

    with its variation, see Eq. :eq:`solidbody-variation`

    ..  math::
        :label: solidbody-variation

        \delta_\boldsymbol{u}(\Pi_{int}) =
            \int_V \frac{\partial \psi}{\partial \boldsymbol{F}} \ dV
        \longrightarrow \boldsymbol{f}_\boldsymbol{u}

    and linearization, see Eq. :eq:`solidbody-linearization`. The right-arrows in
    Eq. :eq:`solidbody-variation` and Eq. :eq:`solidbody-linearization`
    represent the assembly into system scalars, vectors or matrices.

    ..  math::
        :label: solidbody-linearization

        \delta_\boldsymbol{u}\Delta_\boldsymbol{u}(\Pi_{int}) =
            \int_V \delta\boldsymbol{F} :
                \frac{\partial^2 \psi}{
                    \partial \boldsymbol{F}\ \partial \boldsymbol{F}
                } : \Delta\boldsymbol{F}\ dV
        \longrightarrow \boldsymbol{K}_{\boldsymbol{u}\boldsymbol{u}}

    The displacement-based formulation leads to a linearized equation system as given in
    Eq. :eq:`solidbody-final`.

    ..  math::
        :label: solidbody-final

        \boldsymbol{K}_{\boldsymbol{u}\boldsymbol{u}} \cdot \delta \boldsymbol{u} +
            \boldsymbol{f}_\boldsymbol{u} = \boldsymbol{0}

    ..  note::
        This class also supports ``umat`` with mixed-field formulations like
        :class:`~felupe.NearlyIncompressible` or :class:`~felupe.ThreeFieldVariation`.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Cube(n=6)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries, loadcase = fem.dof.uniaxial(field, clamped=True)
    >>>
    >>> umat = fem.NeoHooke(mu=1, bulk=2)
    >>> solid = fem.SolidBody(umat, field)
    >>>
    >>> table = fem.math.linsteps([0, 1], num=5)
    >>> step = fem.Step(
    >>>     items=[solid],
    >>>     ramp={boundaries["move"]: table},
    >>>     boundaries=boundaries,
    >>> )
    >>>
    >>> job = fem.Job(steps=[step]).evaluate()
    >>> ax = solid.imshow("Principal Values of Cauchy Stress")

    See Also
    --------
    felupe.SolidBodyNearlyIncompressible : A (nearly) incompressible solid body with
        methods for the assembly of sparse vectors/matrices.
    """

    def __init__(self, umat, field, statevars=None):
        self.umat = umat
        self.field = field

        self.results = Results(stress=True, elasticity=True)
        self.results.kinematics = self._extract(self.field)

        if statevars is not None:
            self.results.statevars = statevars
        else:
            statevars_shape = (0,)
            if hasattr(umat, "x"):
                statevars_shape = umat.x[-1].shape
            self.results.statevars = np.zeros(
                (
                    *statevars_shape,
                    field.region.quadrature.npoints,
                    field.region.mesh.ncells,
                )
            )

        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

        self.evaluate = Evaluate(
            gradient=self._gradient,
            hessian=self._hessian,
            cauchy_stress=self._cauchy_stress,
            kirchhoff_stress=self._kirchhoff_stress,
        )

        self._area_change = AreaChange()

        self._form = IntegralForm

    def _vector(self, field=None, parallel=False, items=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field

        self.results.stress = self._gradient(field, args=args, kwargs=kwargs)
        self.results.force = self._form(
            fun=self.results.stress[slice(items)],
            v=self.field,
            dV=self.field.region.dV,
        ).assemble(parallel=parallel)

        return self.results.force

    def _matrix(self, field=None, parallel=False, items=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field

        self.results.elasticity = self._hessian(field, args=args, kwargs=kwargs)

        form = self._form(
            fun=self.results.elasticity[slice(items)],
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
        )

        self.results.stiffness_values = form.integrate(
            parallel=parallel, out=self.results.stiffness_values
        )

        self.results.stiffness = form.assemble(values=self.results.stiffness_values)

        return self.results.stiffness

    def _extract(self, field):
        self.field = field
        self.results.kinematics = self.field.extract(out=self.results.kinematics)

        return self.results.kinematics

    def _gradient(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        if "out" in inspect.signature(self.umat.gradient).parameters:
            kwargs["out"] = self.results.gradient

        gradient = self.umat.gradient(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )
        self.results.gradient = gradient[0]

        self.results.stress, self.results._statevars = gradient[:-1], gradient[-1]

        return self.results.stress

    def _hessian(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        if "out" in inspect.signature(self.umat.hessian).parameters:
            kwargs["out"] = self.results.hessian

        self.results.elasticity = self.umat.hessian(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )
        self.results.hessian = self.results.elasticity[0]

        return self.results.elasticity

    def _kirchhoff_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]

        return dot(P, transpose(F))

    def _cauchy_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]
        J = det(F)

        return dot(P, transpose(F)) / J



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._container import MeshContainer
from ._mesh import Mesh


def read(
    filename, file_format=None, cellblock=None, dim=None, merge=False, decimals=None
):
    """Read a mesh from a file using :func:`meshio.read` and create a
    :class:`~felupe.MeshContainer`.

    Parameters
    ----------
    filename : str
        The filename of the mesh file.
    file_format : str or None, optional
        The file format of the mesh file (default is None).
    cellblock : list of int or None, optional
        Read only a subset of the cellblocks from the mesh file (default is None). If
        None, all cell blocks are added to the :class:`~felupe.MeshContainer`.
    dim : int or None, optional
        If provided, the dimension to trim the points array to (default is None). If
        None, the points array is unchanged.
    merge : bool, optional
        Flag to merge duplicate mesh points. This changes the cells arrays of the
        meshes. Default is False.
    decimals : float or None, optional
        Precision decimals for merging duplicated mesh points. Only relevant if
        merge=True. Default is None.

    Returns
    -------
    MeshContainer
        A mesh container created with :func:`meshio.read`.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Rectangle(n=3)
    >>> mesh.write(filename="mesh.xdmf")

    >>> container = fem.mesh.read("mesh.xdmf")
    >>> container
    <felupe mesh container object>
      Number of points: 9
      Number of cells:
        quad: 4

    >>> container.meshes[0]
    <felupe Mesh object>
      Number of points: 9
      Number of cells:
        quad: 4

    See Also
    --------
    meshio.read : Reads an unstructured mesh with added data.
    felupe.Mesh.write : Write the mesh to a file.
    """

    from meshio import read as meshio_read

    m = meshio_read(filename=filename, file_format=file_format)

    if dim is None:
        dim = m.points.shape[1]

    points = m.points[:, :dim]

    if cellblock is None:
        cellblock = slice(None)

    cells = m.cells[cellblock]

    if not isinstance(cells, list):
        cells = [cells]

    if len(cells) > 0:
        meshes = [Mesh(points, c.data, c.type) for c in cells]
    else:
        meshes = [Mesh(points, np.zeros((0, 0), dtype=int), None)]

    return MeshContainer(meshes, merge=merge, decimals=decimals)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt

from scipy.sparse import csr_matrix as sparsematrix


class IntegralFormCartesian:
    r"""Single-field integral form constructed by a function result ``fun``, a test
    field ``v``, differential volumes ``dV`` and optionally a trial field ``u``. For
    both fields ``v`` and ``u`` gradients may be passed by setting ``grad_v`` and
    ``grad_u`` to True (default is False for both fields).

    Arguments
    ---------
    fun : array
        The pre-evaluated function array.
    v : Field
        The test field.
    dV : array
        The differential volumes.
    u : Field, optional
        If a field is passed, a bilinear form is created (default is None).
    grad_v : bool, optional
        Flag to activate the gradient on the test field ``v`` (default is False).
    grad_u : bool, optional
        Flag to activate the gradient on the trial field ``u`` (default is False).

    Notes
    -----

    Linearform
    ~~~~~~~~~~
    A linear form is either defined by the dot product of a given (vector-valued)
    function :math:`\boldsymbol{f}` and the (vector) field :math:`\boldsymbol{v}`

    ..  math::

        L(\boldsymbol{v}) = \int_\Omega \boldsymbol{f} \cdot \boldsymbol{v} ~ dV


    or by the double-dot product of a (matrix-valued) function :math:`\boldsymbol{F}`
    and the gradient of the field values :math:`\boldsymbol{\nabla v}`.

    ..  math::

        L(\boldsymbol{v}) = \int_\Omega \boldsymbol{f} : \boldsymbol{\nabla v} ~ dV

    Bilinearform
    ~~~~~~~~~~~~
    A bilinear form is either defined by the dot products of a given (matrix-valued)
    function :math:`\boldsymbol{f}` and the (vector) fields :math:`\boldsymbol{v}` and
    :math:`\boldsymbol{u}`

    ..  math::

        a(\boldsymbol{v}, \boldsymbol{u}) =
            \int_\Omega \boldsymbol{v} \cdot \boldsymbol{f} \cdot \boldsymbol{u} ~ dV


    or by the double-dot products of a (tensor-valued) function :math:`\boldsymbol{f}`
    and the field values :math:`\boldsymbol{v}` and :math:`\boldsymbol{u}` or their
    gradients :math:`\boldsymbol{\nabla v}` and :math:`\boldsymbol{\nabla u}`.

    ..  math::

        a(\boldsymbol{v}, \boldsymbol{u}) &=
            \int_\Omega \boldsymbol{\nabla v} : \boldsymbol{f} \cdot \boldsymbol{u} ~ dV

        a(\boldsymbol{v}, \boldsymbol{u}) &=
            \int_\Omega \boldsymbol{v} \cdot \boldsymbol{f} : \boldsymbol{\nabla u} ~ dV

        a(\boldsymbol{v}, \boldsymbol{u}) &= \int_\Omega
            \boldsymbol{\nabla v} : \boldsymbol{f} : \boldsymbol{\nabla u} ~ dV

    See Also
    --------
    felupe.IntegralForm : Mixed-field integral form container with methods for integration and assembly.
    felupe.IntegralFormAxisymmetric : An Integral Form for axisymmetric fields.

    """

    def __init__(self, fun, v, dV, u=None, grad_v=False, grad_u=False):
        self.fun = np.ascontiguousarray(fun)
        self.dV = dV

        self.v = v
        self.grad_v = grad_v

        self.u = u
        self.grad_u = grad_u

        # init indices

        # # linear form
        if not self.u:
            self.indices = self.v.indices.ai
            self.shape = self.v.indices.shape

        # # bilinear form
        else:
            cai = self.v.indices.cai
            cbk = self.u.indices.cai

            caibk0 = np.repeat(cai, cbk.shape[1] * self.u.dim)
            caibk1 = np.tile(cbk, (1, cai.shape[1] * self.v.dim, 1)).ravel()

            self.indices = (caibk0, caibk1)
            self.shape = (self.v.indices.shape[0], self.u.indices.shape[0])

    def assemble(self, values=None, parallel=False, out=None):
        "Assembly of sparse region vectors or matrices."

        if values is None:
            values = self.integrate(parallel=parallel, out=out)

        permute = np.append(len(values.shape) - 1, range(len(values.shape) - 1)).astype(
            int
        )

        res = sparsematrix(
            (values.transpose(permute).ravel(), self.indices), shape=self.shape
        )

        return res

    def integrate(self, parallel=False, out=None):
        "Return evaluated (but not assembled) integrals."

        grad_v, grad_u = self.grad_v, self.grad_u
        v, u = self.v, self.u
        dV = self.dV
        fun = self.fun

        # plane strain
        # trim 3d vector-valued functions to the dimension of the field
        function_dimension = len(fun.shape) - 2
        function_is_vector = function_dimension >= 1
        function_is_3d = len(fun) == 3
        field_is_2d = v.dim == 2

        if function_is_vector and function_is_3d and field_is_2d:
            fun = fun[tuple([slice(2)] * function_dimension)]

        if parallel:
            einsum = einsumt
        else:
            einsum = np.einsum

        if not grad_v:
            vb = v.region.h
        else:
            vb = v.region.dhdX

        if u is not None:
            if not grad_u:
                ub = u.region.h
            else:
                ub = u.region.dhdX

        if u is None:
            if not grad_v:
                return einsum(
                    "aqc,...qc,qc->a...c", vb, fun, dV, optimize=True, out=out
                )
            else:
                return einsum(
                    "aJqc,...Jqc,qc->a...c", vb, fun, dV, optimize=True, out=out
                )

        else:
            if not grad_v and not grad_u:
                res = einsum(
                    "aqc,...qc,bqc,qc->a...bc", vb, fun, ub, dV, optimize=True, out=out
                )
                if len(res.shape) == 5:
                    return einsum("aijbc->aibjc", res, out=out)
                else:
                    return res
            elif grad_v and not grad_u:
                return einsum(
                    "aJqc,iJ...qc,bqc,qc->aib...c",
                    vb,
                    fun,
                    ub,
                    dV,
                    optimize=True,
                    out=out,
                )
            elif not grad_v and grad_u:
                return einsum(
                    "a...qc,...kLqc,bLqc,qc->a...bkc",
                    vb,
                    fun,
                    ub,
                    dV,
                    optimize=True,
                    out=out,
                )
            else:  # grad_v and grad_u
                return einsum(
                    "aJqc,iJkLqc,bLqc,qc->aibkc",
                    vb,
                    fun,
                    ub,
                    dV,
                    optimize=True,
                    out=out,
                )
